from unittest import mock

import pytest

# noinspection PyProtectedMember
from alive_progress.core.configuration import Config, ERROR, __style_input_factory, \
    _bool_input_factory, _int_input_factory, create_config, _format_input_factory
from alive_progress.styles.internal import BARS, SPINNERS, THEMES
from alive_progress.utils.terminal import NON_TTY, FULL


@pytest.mark.parametrize('lower, upper, num, expected', [
    (100, 110, 100, 100),
    (100, 110, 110, 110),
    (100, 110, -1, ERROR),
    (100, 110, 111, ERROR),
])
def test_int_input_factory(lower, upper, num, expected):
    func = _int_input_factory(lower, upper)
    assert func(num) == expected


@pytest.mark.parametrize('param, expected', [
    (False, False),
    (0, False),
    ('', False),
    (None, False),
    (True, True),
    (1, True),
    ('asd', True),
])
def test_bool_input_factory(param, expected):
    func = _bool_input_factory()
    assert func(param) == expected


@pytest.mark.parametrize('param, expected', [
    (True, True),
    (False, False),
    (1.2345, True),
    (object(), True),
    (None, False),
    ([], False),
    ('', ''),
    ('any text', 'any text'),
    ('text {{text', 'text {{text'),
    ('text {{apple', 'text {{apple'),
    ('text apple}}', 'text apple}}'),
    ('text {{text}}', 'text {{text}}'),
    ('{kiwi}', '{kiwi}'),
    ('text {kiwi} text', 'text {kiwi} text'),
    ('{mango}', ERROR),
    ('text {mango} text', ERROR),
])
def test_format_input_factory(param, expected):
    func = _format_input_factory('banana apple kiwi')
    assert func(param) == expected


def func_style_123():
    def artifact_super_cool_compiler_assembler_factory():
        pass

    return artifact_super_cool_compiler_assembler_factory


def func_not_style():
    pass


STYLE_1 = func_style_123()
STYLE_2 = func_style_123()
NAMES = dict(name_1=STYLE_1, name_2=STYLE_2)


@pytest.mark.parametrize('param, expected', [
    (STYLE_1, STYLE_1),
    (STYLE_2, STYLE_2),
    ('name_1', STYLE_1),
    ('name_2', STYLE_2),
])
def test_style_input_factory(param, expected):
    test_style_input_factory.__file__ = __file__

    func = __style_input_factory(NAMES, test_style_input_factory,
                                 'artifact_super_cool_compiler_assembler_factory', None)
    assert func(param) == expected


@pytest.mark.parametrize('param', [
    'banana', func_not_style, STYLE_1,
])
def test_style_input_factory_error(param):
    test_style_input_factory_error.__file__ = ''  # simulates a func_style declared elsewhere.

    func = __style_input_factory(NAMES, test_style_input_factory_error,
                                 'artifact_super_cool_compiler_assembler_factory', None)
    assert func(param) is ERROR


@pytest.fixture
def handler():
    yield create_config()


def test_config_creation(handler):
    config = handler()
    assert isinstance(config, Config)


@pytest.fixture(params=[
    (dict(length=9), {}),
    (dict(spinner='pulse'), dict(spinner=SPINNERS['pulse'])),
    (dict(spinner=SPINNERS['pulse']), {}),
    (dict(bar='solid'), dict(bar=BARS['solid'])),
    (dict(bar=BARS['solid']), {}),
    (dict(force_tty=False), dict(force_tty=NON_TTY)),
    (dict(manual=True), {}),
    (dict(enrich_print=False), {}),
    (dict(title_length=20), {}),
    (dict(force_tty=True, manual=True, enrich_print=False, title_length=10), dict(force_tty=FULL)),
    (dict(spinner=None, manual=None), dict(manual=False)),
])
def config_params(request):
    yield request.param


def test_config_global(config_params, handler):
    params, diff = config_params
    expected = dict(params, **diff)
    handler.set_global(**params)
    config = handler()
    assert {k: v for k, v in config._asdict().items() if k in params} == expected


def test_config_local(config_params, handler):
    params, diff = config_params
    expected = dict(params, **diff)
    config = handler(**params)
    assert {k: v for k, v in config._asdict().items() if k in params} == expected


@pytest.fixture(params=[
    dict(length=None),
    dict(length=9999),
    dict(spinner='banana'),
    dict(bar='coolest'),
    dict(unknown=None),
    dict(unknown='nope'),
    dict(theme='rogerio'),
    dict(spinner=SPINNERS['pulse'], bar='oops', unknown=SPINNERS['fish']),
    dict(hey=True),
    dict(length=10, cool='very'),
])
def config_params_error(request):
    yield request.param


def test_config_global_error(config_params_error, handler):
    with pytest.raises(ValueError):
        handler.set_global(**config_params_error)


def test_config_local_error(config_params_error, handler):
    with pytest.raises(ValueError):
        handler(**config_params_error)


@pytest.fixture
def config_params_theme(config_params):
    with mock.patch.dict(THEMES, cool=config_params[0]):
        yield config_params


def test_config_global_theme(config_params_theme, handler):
    params, expected = config_params_theme
    expected = dict(params, **expected)
    handler.set_global(theme='cool')
    config = handler()
    assert {k: getattr(config, k) for k in params} == expected



def write(_text):
    return 0


def flush():
    pass


def _ansi_escape_sequence(_=''):
    def inner(_available=None):
        pass

    return inner


clear_line = _ansi_escape_sequence()
clear_end_line = _ansi_escape_sequence()
clear_end_screen = _ansi_escape_sequence()
hide_cursor = _ansi_escape_sequence()
show_cursor = _ansi_escape_sequence()
factory_cursor_up = lambda _: _ansi_escape_sequence()


def cols():
    return 0  # more details in `alive_progress.tools.sampling#overhead`.


carriage_return = ''



import os
import random
import re
import sys
import time
from collections import namedtuple
from enum import Enum

from .internal import BARS, SPINNERS, THEMES
from ..animations.spinners import scrolling_spinner_factory, sequential_spinner_factory
from ..animations.utils import spinner_player
from ..core.configuration import config_handler
from ..utils.cells import combine_cells, print_cells
from ..utils.terminal import FULL

Show = Enum('Show', 'SPINNERS BARS THEMES')


def showtime(show=Show.SPINNERS, *, fps=None, length=None, pattern=None):
    """Start a show, rendering all styles simultaneously in your screen.

    Args:
        fps (float): the desired frames per second refresh rate
        show (Show): chooses which show will run
        length (int): the bar length, as in configuration options
        pattern (Pattern): to filter objects displayed

    """
    show_funcs = {
        Show.SPINNERS: show_spinners,
        Show.BARS: show_bars,
        Show.THEMES: show_themes,
    }
    assert show in show_funcs, 'Which show do you want? Try Show.SPINNERS for example.'
    show_funcs[show](fps=fps, length=length, pattern=pattern)


Info = namedtuple('Info', 'title descr tech')


def show_spinners(*, fps=None, length=None, pattern=None):
    """Start a spinner show, rendering all styles simultaneously in your screen.

    Args:
        fps (float): the desired frames per second rendition
        length (int): the bar length, as in configuration options
        pattern (Pattern): to filter objects displayed

    """
    selected = _filter(SPINNERS, pattern)
    max_name_length = max(len(s) for s in selected) + 2
    max_natural = max(s.natural for s in selected.values()) + 2
    gens = [_spinner_gen(f'{k:^{max_name_length}}', s, max_natural) for k, s in selected.items()]
    info = Info(
        title=('Spinners', 'including their unknown bar performances'),
        descr=('Spinners generate and run fluid animations, with a plethora of special effects,'
               ' including static frames, scrolling, bouncing, sequential, alongside or delayed!',
               'Each type supports several customization options that allow some very cool tricks,'
               ' so be creative 😜'),
        tech=('Spinners are advanced generators that dynamically output frames to generate some '
              'effect.',
              'These frames are gathered into full cycles, where the spinner yields. This enables'
              ' to mix and match them, without ever breaking animations.',
              'All spinners compile their full animations only once before displaying, so they are'
              ' faaaast!',
              'The spinner compiler brings the super cool `.check()` tool, check it out!',
              'A spinner have a specific "natural" length, and know how to spread its contents over'
              ' any desired space.',))
    _showtime_gen(fps, gens, info, length)


def show_bars(*, fps=None, length=None, pattern=None):
    """Start a bar show, rendering all styles simultaneously in your screen.

    Args:
        fps (float): the desired frames per second rendition
        length (int): the bar length, as in configuration options
        pattern (Pattern): to filter objects displayed

    """
    selected = _filter(BARS, pattern)
    max_name_length = max(len(s) for s in selected) + 2
    gens = [_bar_gen(f'{k:>{max_name_length}}', b) for k, b in selected.items()]
    info = Info(
        title=('Bars', 'playing even their underflow and overflow acts'),
        descr=('A bar can render any percentage with a plethora of effects, including dynamic'
               ' chars, tips, backgrounds, transparent fills, underflows and overflows!',
               'Bars also support some advanced use cases, which do not go only forward...'
               ' Just use manual mode and be creative 😜'),
        tech=('Bars are advanced closures that render percentages with some effect, in a specific'
              ' fixed length.',
              'Bars are not compiled, but support the super cool `.check()` tool, check it out!',
              'Furthermore, bars can render any external spinners inside its own borders.'))
    _showtime_gen(fps, gens, info, length)


def show_themes(*, fps=None, length=None, pattern=None):
    """Start a theme show, rendering all styles simultaneously in your screen.

    Args:
        fps (float): the desired frames per second rendition
        length (int): the bar length, as in configuration options
        pattern (Pattern): to filter objects displayed

    """
    selected = _filter(THEMES, pattern)
    max_name_length = max(len(s) for s in selected) + 2
    themes = {k: config_handler(**v) for k, v in selected.items()}
    max_natural = max(t.spinner.natural for t in themes.values())
    gens = [_theme_gen(f'{k:>{max_name_length}}', c, max_natural) for k, c in themes.items()]
    info = Info(
        title=('Themes', 'featuring their bar, spinner and unknown bar companions'),
        descr=('A theme is an aggregator, it wraps styles that go well together.',),
        tech=('Themes are syntactic sugar, not actually configuration variables (they are elided'
              ' upon usage, only their contents go into the config).',
              'But you can surely customize them, just send any additional config parameters to'
              ' override anything.'))
    _showtime_gen(fps, gens, info, length)


def _filter(source, pattern):
    p = re.compile(pattern or '')
    selected = {k: v for k, v in source.items() if p.search(k)}
    if not selected:
        raise ValueError(f'Nothing was selected with pattern "{pattern}".')
    return selected


# unfortunately these cool infos increase considerably the animations' lengths, and I think it
# would be ugly to just truncate the animated gifs on the README to try to keep them within
# reasonable sizes (I can't afford to increase them, since they're very large already).
# so they will be kind of an easter egg, showing only on users' machines ;)
_INFO = os.getenv('ALIVE_BAR_EXHIBIT_FULL_INFO', '1') != '0'


def _showtime_gen(fps, gens, info, length):
    if not sys.stdout.isatty():
        raise UserWarning('This must be run on a tty connected terminal.')

    def title(t, r=False):
        return scrolling_spinner_factory(t, right=r, wrap=False).pause(center=12),  # 1-tuple.

    def message(m, s=None):
        return scrolling_spinner_factory(f'{m} 👏 ({s})' if s else m, right=False),  # 1-tuple.

    info_spinners = sequential_spinner_factory(
        *(title('Now on stage...')
          + message(*info.title)
          + sum((message(d) for d in info.descr), ())
          + title(f'Technical details')
          + sum((message(d) for d in info.tech), ())
          + title('Enjoy 🤩', True)),
        intermix=False
    )

    # initialize generators, retrieve their line lengths, and create information line.
    fps, length = min(60., max(2., float(fps or 15.))), length or 40
    cols = max(x for _, x in ((next(gen), gen.send((fps, length))) for gen in gens))
    fps_monitor = 'fps: {:.2f}'
    info_player = spinner_player(info_spinners(max(3, cols - len(fps_monitor.format(fps)) - 1)))

    logo = spinner_player(SPINNERS['waves']())
    start, sleep, frame, line_num = time.perf_counter(), 1. / fps, 0, 0
    start, current = start - sleep, start  # simulates the first frame took exactly "sleep" ms.
    FULL.hide_cursor()
    try:
        while True:
            cols, lines = os.get_terminal_size()

            title = 'Welcome to alive-progress!', next(logo)
            print_cells(title, cols)  # line 1.
            FULL.clear_end_line()
            print()

            info = fps_monitor.format(frame / (current - start)), next(info_player)
            print_cells(info, cols)  # line 2.
            FULL.clear_end_line()

            content = [next(gen) for gen in gens]  # always consume gens, to maintain them in sync.
            for line_num, fragments in enumerate(content, 3):
                if line_num > lines:
                    break
                print()
                print_cells(fragments, cols)
                FULL.clear_end_line()

            frame += 1
            current = time.perf_counter()
            time.sleep(max(0., start + frame * sleep - current))
            print(f'\x1b[{line_num - 1}A', end='\r')  # ANSI escape sequence for Cursor Up.
    except KeyboardInterrupt:
        pass
    finally:
        FULL.show_cursor()


def _spinner_gen(name, spinner_factory, max_natural):
    fps, length = yield
    blanks = (' ',) * (max_natural - spinner_factory.natural)
    spinner_gen = exhibit_spinner(spinner_factory())
    unknown_gen = exhibit_spinner(spinner_factory(length))
    yield len(blanks) + spinner_factory.natural + len(name) + length + 4 + 2  # borders/spaces.
    while True:
        yield (blanks, '|', next(spinner_gen), '| ',  # '{1}|{2}| {0} |{3}|'
               name, ' |', next(unknown_gen), '|')


def exhibit_spinner(spinner):
    player = spinner_player(spinner)
    while True:
        yield next(player)


def _bar_gen(name, bar_factory):
    fps, length = yield
    bar_gen = exhibit_bar(bar_factory(length), fps)
    yield len(name) + length + 2 + 1  # borders/spaces.
    while True:
        yield name, ' ', next(bar_gen)[0]  # '{0} {1}'


def exhibit_bar(bar, fps):
    total = int(fps * 5)
    while True:
        # standard use cases, increment till completion, underflow and overflow.
        for s, t in (0, total), (0, int(total * .5)), (int(total * .5), int(total + 1)):
            for pos in range(s, t):
                percent = pos / total
                yield bar(percent), percent
            # generates a small pause in movement between cases, based on the fps.
            percent = t / total
            for _ in range(int(fps * 2)):
                yield bar.end(percent), percent

        # advanced use cases, which do not go always forward.
        factor = random.random() + 1  # a number between 1.0 and 2.0.
        for percent in (1. - x * factor / total for x in range(total)):
            yield bar(percent), percent
        # generates a small giggle, like a real gauge.
        measure, giggle = random.random(), lambda: (random.random() - .5) * .2
        for _ in range(int(fps * 2)):
            percent = measure + giggle()
            yield bar(percent), percent
        # gradually comes to a full stop.
        for t in range(int(fps * 5)):  # smoother stabilization.
            percent = measure + giggle() / 1.04 ** t
            yield bar(percent), percent
        # enjoy the full stop for a while.
        for t in range(int(fps * 2)):
            yield bar(measure), measure


def _theme_gen(name, config, max_natural):
    fps, length = yield
    bar_std = exhibit_bar(config.bar(length), fps)
    bar_unknown = exhibit_bar(config.bar(length, config.unknown), fps)
    blanks = (' ',) * (max_natural - config.spinner.natural)
    spinner = exhibit_spinner(config.spinner())
    yield len(name) + 2 * length + config.spinner.natural + len(blanks) + 4 + 3  # borders/spaces.
    while True:
        yield (name, ' ', next(bar_std)[0], ' ', next(spinner), blanks,  # '{0} {1} {2}{3} {4}'
               ' ', next(bar_unknown)[0])



import pytest

from alive_progress.core.calibration import calibrated_fps, custom_fps


@pytest.mark.parametrize('calibrate, rate, expected', [
    (-5., -5., 10.),
    (-5., 0., 10.),
    (-5., 10., 60.),
    (20., -5., 10.),
    (20., 0., 10.),
    (20., 20., 60.),
    (20., 9., pytest.approx(50., abs=1)),
    (1e15, -5., 10.),
    (1e15, 0., 10.),
    (1e15, 1e15, 60.),
    (1e15, 2e12, pytest.approx(50., abs=1)),
    (1e30, -5., 10.),
    (1e30, 0., 10.),
    (1e30, 1e30, 60.),
    (1e30, 1e25, pytest.approx(50., abs=1)),
])
def test_calibrate(calibrate, rate, expected):
    fps = calibrated_fps(calibrate)
    assert fps(rate) == expected


@pytest.mark.parametrize('rate, expected', [
    (1., 1),
    (10., .1),
    (10. * 60., pytest.approx(.001666666666666)),
    (.1, 10.),
])
def test_custom(rate, expected):
    fps = custom_fps(rate)
    assert fps(rate) == expected



from types import SimpleNamespace

from . import tty


def get(original):
    def cols():
        # it seems both `jupyter notebook` and `jupyter-lab` do not return cols, only 80 default.
        return 120

    def clear_line():
        write(_clear_line)
        flush()

    def clear_end_line(available=None):
        for _ in range(available or 0):
            write(' ')
        flush()

    clear_end_screen = clear_end_line

    # it seems spaces are appropriately handled to not wrap lines.
    _clear_line = f'\r{" " * cols()}\r'

    from .void import factory_cursor_up, hide_cursor, show_cursor  # noqa

    flush, write, carriage_return = original.flush, original.write, original.carriage_return

    return SimpleNamespace(**locals())


BASE = get(tty.BASE)  # support for jupyter notebooks.



import math
from functools import reduce, update_wrapper, wraps
from inspect import signature
from itertools import accumulate, chain, repeat
from typing import Callable

from ..utils.cells import combine_cells, fix_cells, mark_graphemes, split_graphemes


def spinner_player(spinner):
    """Create an infinite generator that plays all cycles of a spinner indefinitely."""

    def inner_play():
        while True:
            yield from spinner()  # instantiates a new cycle in each iteration.

    return inner_play()  # returns an already initiated generator.


def bordered(borders, default):
    """Decorator to include controllable borders in the outputs of a function."""

    def wrapper(fn):
        @wraps(fn)
        def inner_bordered(*args, **kwargs):
            content, right = fn(*args, **kwargs)
            return combine_cells(left_border, content, right or right_border)

        return inner_bordered

    left_border, right_border = extract_fill_graphemes(borders, default)
    return wrapper


def extract_fill_graphemes(text, default):
    """Extract the exact same number of graphemes as default, filling missing ones."""
    text, default = (tuple(split_graphemes(c or '') for c in p) for p in (text or default, default))
    return (mark_graphemes(t or d) for t, d in zip(chain(text, repeat('')), default))


def static_sliding_window(sep, gap, contents, length, right, initial):
    """Implement a sliding window over some content interspersed with a separator.
    It is very efficient, storing data in only one string.

    Note that the implementation is "static" in the sense that the content is pre-
    calculated and maintained static, but actually when the window slides both the
    separator and content seem to be moved.
    Also keep in mind that `right` is for the content, not the window.
    """

    def sliding_window():
        pos = initial
        while True:
            if pos < 0:
                pos += original
            elif pos >= original:
                pos -= original
            yield content[pos:pos + length]
            pos += step

    adjusted_sep = fix_cells((sep * math.ceil(gap / len(sep)))[:gap]) if gap else ''
    content = tuple(chain.from_iterable(chain.from_iterable(zip(repeat(adjusted_sep), contents))))
    original, step = len(content), -1 if right else 1
    assert length <= original, f'window slides inside content, {length} must be <= {original}'
    content += content[:length]
    return sliding_window()


def overlay_sliding_window(background, gap, contents, length, right, initial):
    """Implement a sliding window over some content on top of a background.
    It uses internally a static sliding window, but dynamically swaps the separator
    characters for the background ones, thus making it appear immobile, with the
    contents sliding over it.
    """

    def overlay_window():
        for cells in window:  # pragma: no cover
            yield tuple(b if c == '\0' else c for c, b in zip(cells, background))

    background = (background * math.ceil(length / len(background)))[:length]
    window = static_sliding_window('\0', gap, contents, length, right, initial)
    return overlay_window()


def combinations(nums):
    """Calculate the number of total combinations a few spinners should have together,
    can be used for example with cycles or with frames played at the same time."""

    def lcm(a, b):
        """Calculate the lowest common multiple of two numbers."""
        return a * b // math.gcd(a, b)

    return reduce(lcm, nums)


def split_options(options):
    """Split options that apply to dual elements, either duplicating or splitting."""
    return options if isinstance(options, tuple) else (options, options)


def spread_weighted(actual_length, naturals):
    """Calculate the weighted spreading of the available space for all natural lengths."""
    total = sum(naturals)
    lengths = (actual_length / total * n for n in naturals)
    lengths = [round(x) for x in accumulate(lengths)]  # needs to be resolved.
    lengths = tuple(map(lambda a, b: a - b, lengths, [0] + lengths))
    assert sum(lengths) == actual_length
    return lengths


def fix_signature(func: Callable, source: Callable, skip_n_params: int):
    """Override signature to hide first n parameters."""
    original_doc = func.__doc__
    update_wrapper(func, source)
    if original_doc:
        func.__doc__ = f'{original_doc}\n{func.__doc__}'
    sig = signature(func)
    sig = sig.replace(parameters=tuple(sig.parameters.values())[skip_n_params:])
    func.__signature__ = sig
    return func


def round_even(n):
    """Round a number to the nearest even integer."""
    r = int(n)
    return r + 1 if r & 1 else r



import pytest


@pytest.fixture
def spinner_test():
    def spinner_test(*data):
        def inner_factory(actual_length=None):
            def cycle_data():
                while True:
                    yield from data

            def inner_spinner():
                yield from next(cycle_gen)

            actual_length = actual_length or inner_factory.natural
            nonlocal data
            ratio = int(actual_length / len(data[0][0])) + 1
            data = tuple(tuple(tuple(d * ratio)[:actual_length] for d in cycle) for cycle in data)

            # generate spec info, just like the real one.
            frames = tuple(len(cycle) for cycle in data)
            inner_spinner.__dict__.update(cycles=len(data), length=len(data[0][0]),
                                          frames=frames, total_frames=sum(frames),
                                          natural=inner_factory.natural)

            cycle_gen = cycle_data()
            return inner_spinner

        # shortcut for single char animations.
        data = tuple(tuple(cycle) if isinstance(cycle, str) else cycle for cycle in data)
        # simulate to_cells().
        data = tuple(tuple(tuple(frame) for frame in cycle) for cycle in data)

        inner_factory.natural = len(data[0][0])
        return inner_factory

    yield spinner_test


@pytest.fixture
def show_marks():
    def show_marks(cells):
        return ''.join(x or 'X' for x in cells)

    return show_marks



import math
import threading
import time
from contextlib import contextmanager

from .calibration import calibrated_fps, custom_fps
from .configuration import config_handler
from .hook_manager import buffered_hook_manager, passthrough_hook_manager
from ..utils import terminal
from ..utils.cells import combine_cells, fix_cells, print_cells, to_cells
from ..utils.timing import elapsed_text, eta_text, gen_simple_exponential_smoothing_eta


def alive_bar(total=None, *, calibrate=None, **options):
    """An alive progress bar to keep track of lengthy operations.
    It has a spinner indicator, elapsed time, throughput and ETA.
    When the operation finishes, a receipt is displayed with statistics.

    If the code is executed in a headless environment, ie without a
    connected tty, all features are disabled but the final receipt.

    Another cool feature is that it tracks the actual count in regard of the
    expected count. So it will look different if you send more (or less) than
    expected.

    Also, the bar installs a hook in the system print function that cleans
    any garbage out of the terminal, allowing you to print() effortlessly
    while using the bar.

    Use it like this:

    >>> from alive_progress import alive_bar
    ... with alive_bar(123, 'Title') as bar:  # <-- expected total and bar title
    ...     for item in <iterable>:
    ...         # process item
    ...         bar()  # makes the bar go forward

    The `bar()` method should be called whenever you want the bar to go forward.
    You usually call it in every iteration, but you could do it only when some
    criteria match, depending on what you want to monitor.

    While in a progress bar context, you have two ways to output messages:
      - the usual Python `print()` statement, which will properly clean the line,
        print an enriched message (including the current bar position) and
        continue the bar right below it;
      - the `bar.text('message')` call, which sets a situational message right within
        the bar, usually to display something about the items being processed or the
        phase the processing is in.

    If the bar is over or underused, it will warn you!
    To test all supported scenarios, you can do this:
    >>> for x in 1000, 1500, 700, 0:
    ...    with alive_bar(x) as bar:
    ...        for i in range(1000):
    ...            time.sleep(.005)
    ...            bar()
    Expected results are these (but you have to see them in motion!):
|████████████████████████████████████████| 1000/1000 [100%] in 6.0s (167.93/s)
|██████████████████████████▋⚠            | (!) 1000/1500 [67%] in 6.0s (167.57/s)
|████████████████████████████████████████✗ (!) 1000/700 [143%] in 6.0s (167.96/s)
|████████████████████████████████████████| 1000 in 5.8s (171.91/s)

    Args:
        total (Optional[int]): the total expected count
        calibrate (float): maximum theoretical throughput to calibrate animation speed
        **options: custom configuration options, which override the global configuration:
            title (Optional[str]): an optional, always visible bar title
            length (int): the number of characters to render the animated progress bar
            spinner (Union[None, str, object]): the spinner style to be rendered next to the bar
                accepts a predefined spinner name, a custom spinner factory, or None
            bar (Union[None, str, object]): the bar style to be rendered in known modes
                accepts a predefined bar name, a custom bar factory, or None
            unknown (Union[str, object]): the bar style to be rendered in the unknown mode
                accepts a predefined spinner name, or a custom spinner factory (cannot be None)
            theme (str): a set of matching spinner, bar and unknown
                accepts a predefined theme name
            force_tty (Optional[int|bool]): forces a specific kind of terminal:
                False -> disables animations, keeping only the the final receipt
                True -> enables animations, and auto-detects Jupyter Notebooks!
                None (default) -> auto select, according to the terminal/Jupyter
            disable (bool): if True, completely disables all output, do not install hooks
            manual (bool): set to manually control the bar position
            enrich_print (bool): enriches print() and logging messages with the bar position
            receipt (bool): prints the nice final receipt, disables if False
            receipt_text (bool): set to repeat the last text message in the final receipt
            monitor (bool|str): configures the monitor widget `152/200 [76%]`
                send a string with `{count}`, `{total}` and `{percent}` to customize it
            elapsed (bool|str): configures the elapsed time widget `in 12s`
                send a string with `{elapsed}` to customize it
            stats (bool|str): configures the stats widget `(123.4/s, eta: 12s)`
                send a string with `{rate}` and `{eta}` to customize it
            monitor_end (bool|str): configures the monitor widget within final receipt
                same as monitor, the default format is dynamic, it inherits monitor's one
            elapsed_end (bool|str): configures the elapsed time widget within final receipt
                same as elapsed, the default format is dynamic, it inherits elapsed's one
            stats_end (bool|str): configures the stats widget within final receipt
                send a string with `{rate}` to customize it (no relation to stats)
            title_length (int): fixes the title lengths, or 0 for unlimited
                title will be truncated if longer, and a cool ellipsis "…" will appear at the end
            spinner_length (int): forces the spinner length, or `0` for its natural one
            refresh_secs (int): forces the refresh period, `0` for the reactive visual feedback
            ctrl_c (bool): if False, disables CTRL+C (captures it)
            dual_line (bool): if True, places the text below the bar

    """
    config = config_handler(**options)
    return __alive_bar(config, total, calibrate=calibrate)


@contextmanager
def __alive_bar(config, total=None, *, calibrate=None, _cond=threading.Condition, _sampling=False):
    """Actual alive_bar handler, that exposes internal functions for configuration of
    both normal operation and overhead estimation."""

    if total is not None:
        if not isinstance(total, int):
            raise TypeError(f"integer argument expected, got '{type(total).__name__}'.")
        if total <= 0:
            total = None

    def run(spinner_player, spinner_suffix):
        with cond_refresh:
            while thread:
                event_renderer.wait()
                alive_repr(next(spinner_player), spinner_suffix)
                cond_refresh.wait(1. / fps(run.rate))

    def alive_repr(spinner=None, spinner_suffix=None):
        run.elapsed = time.perf_counter() - run.init
        run.rate = current() / run.elapsed

        fragments = (run.title, bar_repr(run.percent), bar_suffix, spinner, spinner_suffix,
                     monitor(), elapsed(), stats(), *run.text)

        run.last_len = print_cells(fragments, term.cols(), run.last_len, _term=term)
        term.write(run.suffix)
        term.flush()

    def set_text(text=None):
        if text and config.dual_line:
            run.text, run.suffix = ('\n', to_cells(str(text))), term.cursor_up_1.sequence
        else:
            run.text, run.suffix = (to_cells(None if text is None else str(text)),), ''  # 1-tuple.

    def set_title(title=None):
        run.title = _render_title(config, None if title is None else str(title))
        if run.title:
            run.title += (' ',)  # space separator for print_cells.

    if config.manual:
        def bar_handle(percent):  # for manual progress modes.
            hook_manager.flush_buffers()
            run.percent = max(0., float(percent))
            update_hook()
    else:
        def bar_handle(count=1):  # for counting progress modes.
            hook_manager.flush_buffers()
            run.count += max(1, int(count))
            update_hook()

    def start_monitoring(offset=0.):
        term.hide_cursor()
        hook_manager.install()
        bar._handle = bar_handle
        run.init = time.perf_counter() - offset
        event_renderer.set()

    def stop_monitoring():
        term.show_cursor()
        hook_manager.uninstall()
        bar._handle = None
        return time.perf_counter() - run.init

    @contextmanager
    def pause_monitoring():
        event_renderer.clear()
        offset = stop_monitoring()
        alive_repr()
        term.write('\n')
        term.flush()
        try:
            yield
        finally:
            start_monitoring(offset)

    if total or not config.manual:  # we can count items.
        logic_total, current = total, lambda: run.count
        rate_spec, factor, header = 'f', 1.e6, 'on {:d}: '
    else:  # there's only a manual percentage.
        logic_total, current = 1., lambda: run.percent
        rate_spec, factor, header = '%', 1., 'on {:.1%}: '

    if config.refresh_secs:
        fps = custom_fps(config.refresh_secs)
    else:
        fps = calibrated_fps(calibrate or factor)

    run.last_len, run.elapsed, run.count, run.percent = 0, 0., 0, 0.
    run.rate, run.init, run.text, run.title, run.suffix = 0., 0., None, None, None
    thread, event_renderer, cond_refresh = None, threading.Event(), _cond()
    bar_repr, bar_suffix = _create_bars(config)

    if config.disable:
        term, hook_manager = terminal.VOID, passthrough_hook_manager()
    else:
        term = config.force_tty
        hook_manager = buffered_hook_manager(
            header if config.enrich_print else '', current, cond_refresh, term)

    if term.interactive:
        thread = threading.Thread(target=run, args=_create_spinner_player(config))
        thread.daemon = True
        thread.start()

    def monitor_run(f):
        return f.format(count=run.count, total=total, percent=run.percent)

    def monitor_end(f):
        warning = '(!) ' if current() != logic_total else ''
        return f'{warning}{monitor_run(f)}'

    def elapsed_run(f):
        return f.format(elapsed=elapsed_text(run.elapsed, False))

    def elapsed_end(f):
        return f.format(elapsed=elapsed_text(run.elapsed, True))

    if total or config.manual:  # we can track progress and therefore eta.
        def stats_run(f):
            eta = eta_text(gen_eta.send((current(), run.rate)))
            return f.format(rate=run.rate, rate_spec=rate_spec, eta=eta)

        gen_eta = gen_simple_exponential_smoothing_eta(.5, logic_total)
        gen_eta.send(None)
        stats_default = '({rate:.1{rate_spec}}/s, eta: {eta})'
    else:  # unknown progress.
        def stats_run(f):
            return f.format(rate=run.rate, eta='?')

        bar_repr = bar_repr.unknown
        stats_default = '({rate:.1f}/s)'

    def stats_end(f):
        return f.format(rate=run.rate, rate_spec=rate_spec)

    stats_end_default = '({rate:.2{rate_spec}}/s)'

    if total:
        if config.manual:
            def update_hook():
                run.count = math.ceil(run.percent * total)
        else:
            def update_hook():
                run.percent = run.count / total

        monitor_default = '{count}/{total} [{percent:.0%}]'
    else:
        def update_hook():
            pass

        if config.manual:
            monitor_default = '{percent:.0%}'
        else:
            monitor_default = '{count}'
    elapsed_default = 'in {elapsed}'

    monitor = _Widget(monitor_run, config.monitor, monitor_default)
    monitor_end = _Widget(monitor_end, config.monitor_end, monitor.f[:-1])  # space separator.
    elapsed = _Widget(elapsed_run, config.elapsed, elapsed_default)
    elapsed_end = _Widget(elapsed_end, config.elapsed_end, elapsed.f[:-1])  # space separator.
    stats = _Widget(stats_run, config.stats, stats_default)
    stats_end = _Widget(stats_end, config.stats_end, stats_end_default if stats.f[:-1] else '')

    ctrl_c, bar = False, __AliveBarHandle(pause_monitoring, current, set_title, set_text)
    set_text(), set_title()
    start_monitoring()
    try:
        yield bar if not _sampling else locals()
    except KeyboardInterrupt:
        ctrl_c = True
        if config.ctrl_c:
            raise
    finally:
        stop_monitoring()
        if thread:  # lets the internal thread terminate gracefully.
            local_copy, thread = thread, None
            local_copy.join()

        # guarantees last_len is already set...
        if ctrl_c and term.cols() - run.last_len < 2:
            term.cursor_up_1()  # try to not duplicate last line when terminal prints "^C".

        if config.receipt:  # prints the nice but optional final receipt.
            elapsed, stats, monitor = elapsed_end, stats_end, monitor_end
            bar_repr, run.suffix = bar_repr.end, ''
            if not config.receipt_text:
                set_text()
            term.clear_end_screen()
            alive_repr()
            term.write('\n')
        else:
            term.clear_line()
        term.flush()


class _Widget:  # pragma: no cover
    def __init__(self, func, value, default):
        self.func = func
        if isinstance(value, str):
            self.f = value
        elif value:
            self.f = default
        else:
            self.f = ''

        if self.f:
            self.f += ' '  # space separator for print_cells.

    def __call__(self):
        return self.func(self.f)


class _GatedProperty:  # pragma: no cover
    def __set_name__(self, owner, name):
        self.prop = f'_{name}'

    # noinspection PyProtectedMember
    def __get__(self, obj, objtype=None):
        if obj._handle:
            return getattr(obj, self.prop)
        return _noop

    def __set__(self, obj, value):
        raise AttributeError(f"Can't set {self.prop}")


class _GatedAssignProperty(_GatedProperty):  # pragma: no cover
    # noinspection PyProtectedMember
    def __set__(self, obj, value):
        if obj._handle:
            getattr(obj, self.prop)(value)


class __AliveBarHandle:  # pragma: no cover
    pause = _GatedProperty()
    current = _GatedProperty()
    text = _GatedAssignProperty()
    title = _GatedAssignProperty()

    def __init__(self, pause, get_current, set_title, set_text):
        self._handle, self._pause, self._current = None, pause, get_current
        self._title, self._text = set_title, set_text

    # this enables to exchange the __call__ implementation.
    def __call__(self, *args, **kwargs):
        if self._handle:
            self._handle(*args, **kwargs)


def _noop(*_args, **_kwargs):  # pragma: no cover
    pass


def _create_bars(config):  # pragma: no cover
    bar = config.bar
    if bar is None:
        def obj(*_args, **_kwargs):
            pass

        obj.unknown, obj.end = obj, obj
        return obj, ''

    return bar(config.length, config.unknown), ' '


def _create_spinner_player(config):  # pragma: no cover
    spinner = config.spinner
    if spinner is None:
        from itertools import repeat
        return repeat(''), ''

    from ..animations.utils import spinner_player
    return spinner_player(spinner(config.spinner_length)), ' '


def _render_title(config, title=None):
    title, length = to_cells(title or config.title or ''), config.title_length
    if not length:
        return title

    len_title = len(title)
    if len_title <= length:
        # fixed left align implementation for now, there may be more in the future, like
        # other alignments, variable with a maximum size, and even scrolling and bouncing.
        return combine_cells(title, (' ',) * (length - len_title))

    if length == 1:
        return ('…',)

    return combine_cells(fix_cells(title[:length - 1]), ('…',))


def alive_it(it, total=None, *, finalize=None, calibrate=None, **options):
    """New iterator adapter in 2.0, which makes it simpler to monitor any processing.

    Simply wrap your iterable with `alive_it`, and process your items normally!
    >>> from alive_progress import alive_it
    ... import time
    ... items = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    ... for item in alive_it(items):
    ...     time.sleep(.5)  # process item.

    And the bar will just work, it's that simple!

    All `alive_bar` parameters apply as usual, except `total` (which is smarter: if not supplied
    it will be inferred from the iterable using len or length_hint), and `manual` (which can't
    be used in this mode at all).
    To force unknown mode, even when the total would be available, send `total=0`.

    If you want to use other alive_bar's more advanced features, like for instance setting
    situational text messages, you can assign it to a variable! And send a `finalize` closure
    to set the final receipt title and/or text!

    >>> from alive_progress import alive_it
    ... bar = alive_it(items):
    ... for item in bar:
    ...     bar.text(f'Wow, it works! Item: {item}')
    ...     # process item.

    Args:
        it (iterable): the input iterable to be processed
        total: same as alive_bar
        finalize: a function to be called when the bar is going to finalize
        calibrate: same as alive_bar
        options: same as alive_bar

    See Also:
        alive_bar

    Returns:
        Generator

    """
    config = config_handler(**options)
    if config.manual:
        raise UserWarning("Manual mode can't be used in iterator adapter.")

    if total is None and hasattr(it, '__len__'):
        total = len(it)
    it = iter(it)
    if total is None and hasattr(it, '__length_hint__'):
        total = it.__length_hint__()
    return __AliveBarIteratorAdapter(it, finalize, __alive_bar(config, total, calibrate=calibrate))


class __AliveBarIteratorAdapter:
    def __init__(self, it, finalize, inner_bar):
        self._it, self._finalize, self._inner_bar = it, finalize, inner_bar

    def __iter__(self):
        if '_bar' in self.__dict__:  # this iterator has already initiated.
            return

        with self._inner_bar as self._bar:
            del self._inner_bar
            for item in self._it:
                yield item
                self._bar()
            if self._finalize:
                self._finalize(self._bar)

    def __call__(self, *args, **kwargs):
        raise UserWarning('The bar position is controlled automatically with `alive_it`.')

    def __getattr__(self, item):
        # makes this adapter work as the real bar.
        return getattr(self._bar, item)

    def __setattr__(self, key, value):
        # makes this adapter work as the real bar.
        if '_bar' in self.__dict__:
            return setattr(self._bar, key, value)
        return super().__setattr__(key, value)



import logging
import sys
from contextlib import contextmanager
from threading import Condition
from unittest import mock

import click
import pytest

from alive_progress.core.hook_manager import buffered_hook_manager
from alive_progress.utils.terminal import FULL, VOID


@contextmanager
def hook(hook_manager):
    hook_manager.install()
    yield
    hook_manager.uninstall()


@pytest.fixture(params=[
    ('ok', 'nice 35! ok\n'),
    ('ok  ', 'nice 35! ok\n'),
    ('  ok', 'nice 35!   ok\n'),
    ('  ok  ', 'nice 35!   ok\n'),
])
def print_data(request):
    yield request.param


def test_hook_manager_captures_stdout(print_data, capsys):
    out, expected = print_data
    hook_manager = buffered_hook_manager('nice {}! ', lambda: 35, Condition(), FULL)
    with hook(hook_manager):
        print(out)
    assert capsys.readouterr().out == expected


def test_hook_manager_captures_bytes_stdout(print_data, capsys):
    out, expected = print_data
    hook_manager = buffered_hook_manager('nice {}! ', lambda: 35, Condition(), FULL)
    with hook(hook_manager):
        click.echo(out)
    assert capsys.readouterr().out == expected


# I couldn't make this work yet, there's some weird interaction
# between my hook and the pytest one...
def _hook_manager_captures_logging(capsys):
    import sys
    logging.basicConfig(stream=sys.stderr)
    logger = logging.getLogger('?name?')

    hook_manager = buffered_hook_manager('nice {}! ', lambda: 35, Condition(), FULL)
    with hook(hook_manager):
        logger.error('oops')
    assert capsys.readouterr().err == 'nice 35! ERROR:?name?:oops\n'


def test_hook_manager_captures_multiple_lines(capsys):
    hook_manager = buffered_hook_manager('nice {}! ', lambda: 35, Condition(), FULL)
    with hook(hook_manager):
        print('ok1\nok2')
    assert capsys.readouterr().out == 'nice 35! ok1\n         ok2\n'


def test_hook_manager_can_be_disabled(capsys):
    hook_manager = buffered_hook_manager('', None, Condition(), FULL)
    with hook(hook_manager):
        print('ok')
    assert capsys.readouterr().out == 'ok\n'


def test_hook_manager_flush(capsys):
    hook_manager = buffered_hook_manager('', None, Condition(), FULL)
    with hook(hook_manager):
        print('ok', end='')
        assert capsys.readouterr().out == ''
        hook_manager.flush_buffers()
        assert capsys.readouterr().out == 'ok\n'

    # after used, the buffers should be empty
    hook_manager.flush_buffers()
    assert capsys.readouterr().out == ''


def test_hook_manager_do_clear_line_on_stdout():
    hook_manager = buffered_hook_manager('', None, Condition(), VOID)
    m_clear = mock.Mock()
    with hook(hook_manager), mock.patch.dict(VOID.__dict__, clear_line=m_clear):
        print('some')
    m_clear.assert_called()


def test_hook_manager_do_not_flicker_screen_when_logging():
    logging.basicConfig()
    logger = logging.getLogger()

    hook_manager = buffered_hook_manager('', None, Condition(), VOID)
    m_clear = mock.Mock()
    with hook(hook_manager), mock.patch.dict(VOID.__dict__, clear_line=m_clear):
        logger.error('oops')
    m_clear.assert_not_called()


@pytest.fixture
def handlers():
    handlers = (logging.StreamHandler(sys.stderr),
                logging.StreamHandler(sys.stdout),
                logging.FileHandler('/dev/null', delay=True))
    [logging.root.addHandler(h) for h in handlers]
    yield handlers
    [logging.root.removeHandler(h) for h in handlers]


def test_install(handlers):
    hook_manager = buffered_hook_manager('', None, Condition(), FULL)
    with mock.patch('alive_progress.core.hook_manager._set_stream') as mock_set_stream:
        hook_manager.install()
    mock_set_stream.assert_has_calls(tuple(mock.call(h, mock.ANY) for h in handlers))


def test_uninstall(handlers):
    hook_manager = buffered_hook_manager('', None, Condition(), FULL)
    with mock.patch('alive_progress.core.hook_manager._set_stream') as mock_set_stream:
        hook_manager.install()
        hook_manager.uninstall()
    mock_set_stream.assert_has_calls(tuple(mock.call(h, mock.ANY) for h in handlers))



def elapsed_text(seconds, precise):
    seconds = round(seconds, 1 if precise else 0)
    if seconds < 60.:
        return '{:{}f}s'.format(seconds, .1 if precise else .0)

    minutes, seconds = divmod(seconds, 60.)
    if minutes < 60.:
        return '{:.0f}:{:0{}f}'.format(minutes, seconds, 4.1 if precise else 2.0)

    hours, minutes = divmod(minutes, 60.)
    return '{:.0f}:{:02.0f}:{:0{}f}'.format(hours, minutes, seconds, 4.1 if precise else 2.0)


def eta_text(eta):
    if eta is None or eta < 0.:
        return '-'
    return elapsed_text(eta, False)


def simple_eta(logic_total, pos, rate):
    return (logic_total - pos) / rate


def gen_simple_exponential_smoothing_eta(alfa, logic_total):
    """Implements a generator with a simple exponential smoothing of the
    eta time series.
    Given alfa and y_hat (t-1), we can calculate the next y_hat:
        y_hat = alfa * y + (1 - alfa) * y_hat
        y_hat = alfa * y + y_hat - alfa * y_hat
        y_hat = y_hat + alfa * (y - y_hat)

    Args:
        alfa (float): the smoothing coefficient
        logic_total (float):

    Returns:

    """
    pos = rate = None
    while not rate:
        pos, rate = yield
    y_hat = simple_eta(logic_total, pos, rate)
    while True:
        temp, rate = yield y_hat
        if temp == pos:  # reduce numbers bouncing around.
            continue
        pos = temp
        y = simple_eta(logic_total, pos, rate)
        y_hat += alfa * (y - y_hat)



from .bars import bar_factory
from .spinners import alongside_spinner_factory, bouncing_spinner_factory, \
    delayed_spinner_factory, frame_spinner_factory, scrolling_spinner_factory, \
    sequential_spinner_factory
from .utils import spinner_player

__all__ = (
    'bar_factory', 'spinner_player',
    'frame_spinner_factory', 'scrolling_spinner_factory', 'bouncing_spinner_factory',
    'sequential_spinner_factory', 'alongside_spinner_factory', 'delayed_spinner_factory',
)



from .exhibit import Show, show_bars, show_spinners, show_themes, showtime
from .internal import BARS, SPINNERS, THEMES

__all__ = (
    'BARS', 'SPINNERS', 'THEMES', 'Show', 'showtime', 'show_spinners', 'show_bars', 'show_themes',
)

# I want to create some kind of extension point here. Then users would be able to create
# and share their custom styles with the community! Ideally it should be simple to distribute
# them, don't know for sure how. As a last resort, I could include them all here in this package,
# via the usual fork / pull request contribution, and I'll dynamically find and make them usable.



import math
import time

from about_time import about_time

from .utils import bordered, extract_fill_graphemes, fix_signature, spinner_player
from ..utils.cells import VS_15, combine_cells, fix_cells, has_wide, is_wide, join_cells, \
    mark_graphemes, split_graphemes, strip_marks, to_cells
from ..utils.colors import BLUE, BLUE_BOLD, CYAN, DIM, GREEN, ORANGE, ORANGE_BOLD, RED, YELLOW_BOLD
from ..utils.terminal import FULL


def bar_factory(chars=None, *, tip=None, background=None, borders=None, errors=None):
    """Create a factory of a bar with the given styling parameters.
    Supports unicode grapheme clusters and emoji chars (those that has length one but when on
    screen occupies two cells).

    Now supports transparent fills! Just send a tip, and leave `chars` as None.
    Also tips are now considered for the 100%, which means it smoothly enters and exits the
    frame to get to 100%!! The effect is super cool, use a multi-char tip to see.

    Args:
        chars (Optional[str]): the sequence of increasing glyphs to fill the bar
            can be None for a transparent fill, unless tip is also None.
        tip (Optional[str): the tip in front of the bar
            can be None, unless chars is also None.
        background (Optional[str]): the pattern to be used underneath the bar
        borders (Optional[Union[str, Tuple[str, str]]): the pattern or patterns to be used
            before and after the bar
        errors (Optional[Union[str, Tuple[str, str]]): the pattern or patterns to be used
            when an underflow or overflow occurs

    Returns:
        a styled bar factory

    """

    @bar_controller
    def inner_bar_factory(length, spinner_factory=None):
        if chars:
            if is_wide(chars[-1]):  # previous chars can be anything.
                def fill_style(complete, filling):  # wide chars fill.
                    odd = bool(complete % 2)
                    fill = (None,) if odd != bool(filling) else ()  # odd XOR filling.
                    fill += (chars[-1], None) * int(complete / 2)  # already marked wide chars.
                    if filling and odd:
                        fill += mark_graphemes((chars[filling - 1],))
                    return fill
            else:  # previous chars cannot be wide.
                def fill_style(complete, filling):  # narrow chars fill.
                    fill = (chars[-1],) * complete  # unneeded marks here.
                    if filling:
                        fill += (chars[filling - 1],)  # no widies here.
                    return fill
        else:
            def fill_style(complete, filling):  # invisible fill.
                return fix_cells(padding[:complete + bool(filling)])

        def running(fill):
            return None, (fix_cells(padding[len(fill) + len_tip:]),)  # this is a 1-tuple.

        def ended(fill):
            border = None if len(fill) + len(underflow) <= length else underflow
            texts = *(() if border else (underflow,)), blanks
            return border, texts

        @bordered(borders, '||')
        def draw_known(apply_state, percent):
            virtual_fill = round(virtual_length * max(0., min(1., percent)))
            fill = fill_style(*divmod(virtual_fill, num_graphemes))
            border, texts = apply_state(fill)
            border = overflow if percent > 1. else None if percent == 1. else border
            return fix_cells(combine_cells(fill, tip, *texts)[len_tip:length + len_tip]), border

        if spinner_factory:
            @bordered(borders, '||')
            def draw_unknown(_percent=None):
                return next(player), None

            player = spinner_player(spinner_factory(length))
        else:
            draw_unknown = None

        padding = (' ',) * len_tip + background * math.ceil((length + len_tip) / len(background))
        virtual_length, blanks = num_graphemes * (length + len_tip), (' ',) * length
        return draw_known, running, ended, draw_unknown

    assert chars or tip, 'tip is mandatory for transparent bars'
    assert not (chars and not is_wide(chars[-1]) and has_wide(chars)), \
        'cannot use grapheme with a narrow last char'

    chars = split_graphemes(chars or '')  # the only one not yet marked.
    tip, background = (to_cells(x) for x in (tip, background or ' '))
    underflow, overflow = extract_fill_graphemes(errors, (f'⚠{VS_15}', f'✗{VS_15}'))
    num_graphemes, len_tip = len(chars) or 1, len(tip)
    return inner_bar_factory


def bar_controller(inner_bar_factory):
    def bar_assembler_factory(length, spinner_factory=None):
        """Assembles this bar into an actual bar renderer.

        Args:
            length (int): the bar rendition length (excluding the borders)
            spinner_factory (Optional[spinner_factory]): enable this bar to act in unknown mode

        Returns:
            a bar renderer

        """
        with about_time() as t_compile:
            draw_known, running, ended, draw_unknown = inner_bar_factory(length, spinner_factory)

        def draw(percent):
            return draw_known(running, percent)

        def draw_end(percent):
            return draw_known(ended, percent)

        def bar_check(*args, **kwargs):  # pragma: no cover
            return check(draw, t_compile, *args, **kwargs)

        draw.__dict__.update(
            end=draw_end, unknown=draw_unknown,
            check=fix_signature(bar_check, check, 2),
        )

        if draw_unknown:
            def draw_unknown_end(_percent=None):
                return draw_known(ended, 1.)

            draw_unknown.end = draw_unknown_end

        return draw

    def compile_and_check(*args, **kwargs):  # pragma: no cover
        """Compile this bar factory at some length, and..."""
        # since a bar does not have a natural length, I have to choose one...
        bar_assembler_factory(40).check(*args, **kwargs)  # noqa

    bar_assembler_factory.__dict__.update(
        check=fix_signature(compile_and_check, check, 2),
    )

    return bar_assembler_factory


def check(bar, t_compile, verbosity=0, *, steps=20):  # noqa  # pragma: no cover
    """Check the data, codepoints, and even the animation of this bar.

    Args:
        verbosity (int): change the verbosity level
                         0 for brief data only (default)
                               /                 \\
                              /           3 to include animation
                             /                      \\
            1 to unfold bar data   ----------   4 to unfold bar data
                            |                        |
            2 to reveal codepoints   --------   5 to reveal codepoints
        steps (int): number of steps to display the bar progress

    """
    verbosity = max(0, min(5, verbosity or 0))
    if verbosity in (1, 2, 4, 5):
        render_data(bar, verbosity in (2, 5), steps)
    else:
        spec_data(bar)  # spec_data here displays only brief data, shown only if not full.

    duration = t_compile.duration_human.replace('us', 'µs')
    print(f'\nBar style compiled in: {GREEN(duration)}')
    print(f'(call {HELP_MSG[verbosity]})')

    if verbosity in (3, 4, 5):
        animate(bar)


SECTION = ORANGE_BOLD
CHECK = lambda p: f'{BLUE(f".{check.__name__}(")}{BLUE_BOLD(p)}{BLUE(")")}'
HELP_MSG = {
    0: f'{CHECK(1)} to unfold bar data, or {CHECK(3)} to include animation',
    1: f'{CHECK(2)} to reveal codepoints, or {CHECK(4)} to include animation,'
       f' or {CHECK(0)} to fold up bar data',
    2: f'{CHECK(5)} to include animation, or {CHECK(1)} to hide codepoints',
    3: f'{CHECK(4)} to unfold bar data, or {CHECK(0)} to omit animation',
    4: f'{CHECK(5)} to reveal codepoints, or {CHECK(1)} to omit animation,'
       f' or {CHECK(3)} to fold up bar data',
    5: f'{CHECK(2)} to omit animation, or {CHECK(4)} to hide codepoints',
}


def spec_data(bar):  # pragma: no cover
    print(f'\n{SECTION("Brief bar data")}')
    info = lambda field, p, b: f'{YELLOW_BOLD(field, "<11")}: {" ".join(bar_repr(b, p)[1:])}'
    print('\n'.join(info(n, p, bar) for n, p in (
        ('starting', 0.), ('in progress', .5), ('completed', 1.), ('overflow', 1.2)
    )))
    print(info('underflow', .5, bar.end))


def format_codepoints(frame):  # pragma: no cover
    codes = '|'.join((ORANGE if is_wide(g) else BLUE)(
        ' '.join(hex(ord(c)).replace('0x', '') for c in g)) for g in frame)
    return f" -> {RED(sum(len(fragment) for fragment in frame))}:[{codes}]"


def render_data(bar, show_codepoints, steps):  # pragma: no cover
    print(f'\n{SECTION("Full bar data")}', end='')
    codepoints = format_codepoints if show_codepoints else lambda _: ''
    for name, b in ('in progress', bar), ('completed', bar.end):
        print(f'\n{name}')
        for p in (x / steps for x in range(steps + 2)):
            frame, joined, perc = bar_repr(b, p)
            print(joined, perc, codepoints(frame))


def bar_repr(bar, p):  # pragma: no cover
    frame = tuple(strip_marks(bar(p)))
    return frame, ''.join(frame), DIM(f'{p:6.1%}')


def animate(bar):  # pragma: no cover
    print(f'\n{SECTION("Animation")}')
    from ..styles.exhibit import exhibit_bar
    bar_gen = exhibit_bar(bar, 15)
    FULL.hide_cursor()
    try:
        while True:
            rendition, percent = next(bar_gen)
            print(f'\r{join_cells(rendition)}', CYAN(max(0., percent), "6.1%"))
            print(DIM('(press CTRL+C to stop)'), end='')
            FULL.clear_end_line()
            time.sleep(1 / 15)
            FULL.cursor_up_1()
    except KeyboardInterrupt:
        pass
    finally:
        FULL.show_cursor()



import shutil
import sys
from types import SimpleNamespace


def get(original):
    write = original.write
    flush = original.flush

    def cols():
        # more resilient one, although 7x slower than os' one.
        return shutil.get_terminal_size()[0]

    def _ansi_escape_sequence(code, param=''):
        def inner(_available=None):  # because of jupyter.
            write(inner.sequence)

        inner.sequence = f'\x1b[{param}{code}'
        return inner

    def factory_cursor_up(num):
        return _ansi_escape_sequence('A', num)  # sends cursor up: CSI {x}A.

    clear_line = _ansi_escape_sequence('2K\r')  # clears the entire line: CSI n K -> with n=2.
    clear_end_line = _ansi_escape_sequence('K')  # clears line from cursor: CSI K.
    clear_end_screen = _ansi_escape_sequence('J')  # clears screen from cursor: CSI J.
    hide_cursor = _ansi_escape_sequence('?25l')  # hides the cursor: CSI ? 25 l.
    show_cursor = _ansi_escape_sequence('?25h')  # shows the cursor: CSI ? 25 h.
    carriage_return = '\r'

    return SimpleNamespace(**locals())


BASE = get(sys.stdout)  # support for jupyter notebooks.



import sys

import pytest

from alive_progress.utils.cells import to_cells, print_cells
from alive_progress.utils.terminal import tty


@pytest.mark.parametrize('text, expected', [
    ('', ''),
    (None, ''),
    ('a text', 'a text'),
    ('\n', ' '),
    (' \n ', '   '),
    ('\n \n', '   '),
    ('\r', ' '),
    (' \r ', '   '),
    ('\r \n', '   '),
    ('asd\n', 'asd '),
    ('\nasd', ' asd'),
    ('asd1\nasd2', 'asd1 asd2'),
    ('asd1 \nasd2', 'asd1  asd2'),
    ('asd1 \r\nasd2', 'asd1   asd2'),
    ('\nasd1\n \r \nasd2\r', ' asd1     asd2 '),
])
def test_sanitize_text_normal_chars(text, expected, show_marks):
    result = to_cells(text)
    assert show_marks(result) == expected


@pytest.mark.parametrize('text, expected', [
    ('😺', '😺X'),
    ('\n😺', ' 😺X'),
    ('😺 \n 😺', '😺X   😺X'),
    ('\n 😺\n😺', '  😺X 😺X'),
    ('asd😺\n', 'asd😺X '),
    ('😺\nasd', '😺X asd'),
    ('asd1\rasd2😺', 'asd1 asd2😺X'),
    ('\nasd1😺\n😺\n\rasd2\r', ' asd1😺X 😺X  asd2 '),
])
def test_sanitize_text_wide_chars(text, expected, show_marks):
    result = to_cells(text)
    assert show_marks(result) == expected


@pytest.mark.parametrize('text, expected', [
    ('ok', 'ok'),
    ('😺', '😺X'),
    ('😺😺', '😺X😺X'),
    ('😺ok😺', '😺Xok😺X'),
])
def test_sanitize_text_double(text, expected, show_marks):
    result = to_cells(text)
    assert show_marks(result) == expected


@pytest.mark.parametrize('fragments, cols, ret, expected', [
    (('ok', ' ', '1'), 10, 4, '\rok 1'),
    (('ok', ' ', '1'), 4, 4, '\rok 1'),
    (('ok', ' ', '1'), 3, 3, '\rok '),
    (('ok', '1'), 3, 3, '\rok1'),
    (('ok', '1'), 1, 1, '\ro'),
    (('rogerio', '\n', '1'), 3, 1, '\rrog\x1b[K\n1'),
    (('rogerio', '\n', '12345'), 3, 3, '\rrog\x1b[K\n123'),
])
def test_print_cells(fragments, cols, ret, expected, capsys):
    term = tty.get(sys.stdout)
    assert print_cells(fragments, cols, _term=term) == ret
    term.flush()
    assert capsys.readouterr().out == expected


def test_print_cells_clear(capsys):
    term = tty.get(sys.stdout)
    msg = 'loooong'
    assert print_cells((msg,), 100, 8, _term=term) == len(msg)
    term.flush()
    assert capsys.readouterr().out == f'\r{msg}\x1b[K'



from unittest import mock

import pytest

from alive_progress.utils.timing import elapsed_text, eta_text, \
    gen_simple_exponential_smoothing_eta, simple_eta


@pytest.mark.parametrize('elapsed, precise, expected', [
    (1.4, False, '1s'),
    (1.4, True, '1.4s'),
    (1.45, True, '1.4s'),
    (1.5, False, '2s'),
    (1.5, True, '1.5s'),
    (1.55, True, '1.6s'),
    (61.4, False, '1:01'),
    (61.4, True, '1:01.4'),
    (119.5, False, '2:00'),
    (119.5, True, '1:59.5'),
    (119.95, True, '2:00.0'),
    (120.1, False, '2:00'),
    (120.1, True, '2:00.1'),
    (4000, False, '1:06:40'),
    (4000, True, '1:06:40.0'),
])
def test_to_elapsed_text(elapsed, precise, expected):
    assert elapsed_text(elapsed, precise) == expected


@pytest.mark.parametrize('eta, expected', [
    (None, '-'),
    (-.1, '-'),
    (-1000., '-'),
    (0., '0s'),
    (10., '10s'),
])
def test_to_eta_text(eta, expected):
    assert eta_text(eta) == expected


def test_simple_eta():
    assert simple_eta(10, 2, 5.) == pytest.approx(1.6)


def test_gen_simple_exponential_smoothing_eta():
    with mock.patch('alive_progress.utils.timing.simple_eta') as m_simple_eta:
        m_simple_eta.side_effect = lambda _1, _2, r: r

        data = (  # according to the study of simple_exponential_smoothing in docs directory.
            (88., 88.),
            (75., 81.5),
            (60., 70.75),
            (75., 72.875),
            (56., 64.4375),
            (41., 52.71875),
            (51., 51.859375),
            (23., 37.4296875),
            (22., 29.71484375),
            (17., 23.357421875),
            (12., 17.6787109375),
            (20., 18.83935546875),
            (9., 13.919677734375),
            (5., 9.4598388671875),
            (3., 6.22991943359375),
        )
        gen_eta = gen_simple_exponential_smoothing_eta(.5, 0)
        next(gen_eta)
        for i, (y, ses) in enumerate(data):
            assert gen_eta.send((i, y)) == pytest.approx(ses)  # test actual algorithm.
            assert gen_eta.send((i, None)) == pytest.approx(ses)  # test cached if same position.



import logging
import time
from typing import NamedTuple

from .sampling import OVERHEAD_SAMPLING
from .utils import toolkit
from .. import alive_bar
from ..styles import BARS
from ..utils.colors import BOLD, ORANGE_IT


class Case(NamedTuple):
    name: str = None
    count: int = None
    config: dict = None
    done: bool = None
    hooks: bool = None
    title: str = None


def title(text):
    print(f'=== {BOLD.color_code}{ORANGE_IT(text)} ===')


cases = [
    Case(title='Definite/unknown modes'),
    Case('Normal+total', 1000, dict(total=1000)),
    Case('Underflow+total', 800, dict(total=1200)),
    Case('Overflow+total', 1200, dict(total=800)),
    Case('Unknown', 1000, dict(total=0)),

    Case(title='Manual modes'),
    Case('Normal+total+manual', 1000, dict(total=1000, manual=True)),
    Case('Underflow+total+manual', 800, dict(total=1200, manual=True)),
    Case('Overflow+total+manual', 1200, dict(total=800, manual=True)),
    Case('Unknown+manual', 1000, dict(total=0, manual=True)),

    Case(title='Print and Logging hooks'),
    Case('Simultaneous', 1000, dict(total=1000), hooks=True),

    # title('Quantifying mode')  # soon, quantifying mode...
    # ('Calculating auto', 1000, dict(total=..., manual=False)),
    # ('Calculating manual', 1000, dict(total=..., manual=True)),

    Case(title='Display features'),
    Case('Styles', 1000, dict(total=1000, bar='halloween', spinner='loving'))
]
features = [dict(total=1000, bar=bar, spinner='loving') for bar in BARS]
cases += [Case(name.capitalize(), 1000, {**features[i % len(BARS)], **config}, done=True)
          for i, (name, config) in enumerate(OVERHEAD_SAMPLING, 1)]


def demo(sleep=None):
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    for case in cases:
        if case.title:
            print()
            title(case.title)
            continue
        manual, total = (case.config.get(x) for x in ('manual', 'total'))
        with alive_bar(title_length=22, title=case.name, **case.config) as bar:
            # bar.text('Quantifying...')
            # time.sleep(0)
            bar.text('Processing...')
            time.sleep(0)
            # bar.reset(total)
            for i in range(1, case.count + 1):
                time.sleep(sleep or .003)
                if manual:
                    bar((float(i) / (total or case.count)))
                else:
                    bar()
                if case.hooks and i:
                    if i == 400:
                        print('nice hassle-free print hook!')  # tests hook manager.
                    elif i == 800:
                        logger.info('and even logging hook!!!')  # tests hook manager.
            if case.done:
                bar.text('Ok, done!')


if __name__ == '__main__':
    parser, run = toolkit('Demonstrates alive-progress, showcasing several common scenarios.')
    parser.add_argument('sleep', type=float, nargs='?', help='the sleep time (default=.003)')

    run(demo)



import os.path
import urllib.request
from itertools import zip_longest

from .utils import toolkit
from ..utils.cells import split_graphemes
from ..utils.colors import GREEN, ORANGE, RED

CACHE = '.unicode_cache'


def validate_unicode_breaks(uver=None, show_all=False, cache=True):
    # validate unicode grapheme clusters detection.
    # this downloads the specs directly from unicode.org and caches it locally.
    # document: https://unicode.org/reports/tr51/

    latest = f'{CACHE}/latest'
    if not uver and cache and os.path.exists(latest):
        with open(latest) as f:
            uver = f.read()
        print('using version "latest" as:', uver)

    file = f'{CACHE}/emoji-test_{uver}.txt'
    if cache and os.path.exists(file):
        print('loading cached:', file)
        with open(file) as f:
            data = f.read()
    else:
        url = f'https://www.unicode.org/Public/emoji/{uver or "latest"}/emoji-test.txt'
        print('downloading:', url)

        try:
            req = urllib.request.urlopen(url)
        except OSError as e:
            print(RED('Download error:'), e)
            return

        os.makedirs(os.path.dirname(file), exist_ok=True)
        if not uver:
            new_url = req.geturl()
            uver = new_url.split('/')[5]
            print('saving version "latest" as:', uver)
            with open(latest, 'w') as f:
                f.write(uver)
            file = f'{CACHE}/emoji-test_{uver}.txt'

        print('saving:', file)
        data = req.read().decode('utf8')
        with open(file, 'w') as f:
            f.write(data)

    def where():
        nonlocal groups
        if any(groups):
            print('\n'.join(g for g in groups if g))
        groups = [None, None]

    def expect(*chars):
        nonlocal errors, total
        text = ''.join(chars)
        actual = split_graphemes(text)
        error = actual != chars
        total += 1
        errors += error
        if error or show_all:
            where()
            codes = '|'.join((GREEN if a == c else RED)(
                ' '.join(hex(ord(c)).replace('0x', '') for c in a) if a else '-'
            ) for a, c in zip_longest(actual, chars))
            small_name = name.replace(' skin tone', '').replace(' hair', '')
            small_status = ''.join(x[0] for x in status.split('-'))
            a_len, c_len = len(actual), len(chars)
            size = f'{GREEN(a_len)} ==' if a_len == c_len else f'{RED(a_len)} !='
            print(f' {char}   {text.replace(char, "X"):>3}: {size} {c_len} -> '
                  f'|{codes}| {ORANGE(small_status)} {small_name}')

    groups, total, errors = [None, None], 0, 0
    for line in filter(None, data.splitlines()):
        if line.startswith('#'):
            if line.startswith('# group:'):
                groups[0] = line.split()[-1]
            elif line.startswith('# subgroup:'):
                groups[1] = f'  - {line.split()[-1]}'
            continue

        p1, p2 = (p.split() for p in line.split(';'))
        status, name = p2[0], ' '.join(p2[4:])
        char = ''.join(chr(int(x, 16)) for x in p1)
        expect(char)
        expect(char, char)
        expect('a', char, 'a')
        expect('a', 'a', char)
        expect(char, 'a', 'a')

    print(f'\nerrors   : {errors / total:6.2%} [{errors}/{total}]')
    print(f'successes: {1 - errors / total:6.2%} [{total - errors}/{total}]')


def find_groups(data, max_diff):
    """Group some numbers with a maximum difference between them.
    I've used to try to fix the current grapheme break error.

    Using version unicode 13.1:
        Component
          - skin-tone
         🏻    XX: 1 != 2 -> |1f3fb 1f3fb|-| c light
         🏻   aXa: 2 != 3 -> |61 1f3fb|61|-| c light
         🏻   aaX: 2 != 3 -> |61|61 1f3fb|-| c light
         🏼    XX: 1 != 2 -> |1f3fc 1f3fc|-| c medium-light
         🏼   aXa: 2 != 3 -> |61 1f3fc|61|-| c medium-light
         🏼   aaX: 2 != 3 -> |61|61 1f3fc|-| c medium-light
         🏽    XX: 1 != 2 -> |1f3fd 1f3fd|-| c medium
         🏽   aXa: 2 != 3 -> |61 1f3fd|61|-| c medium
         🏽   aaX: 2 != 3 -> |61|61 1f3fd|-| c medium
         🏾    XX: 1 != 2 -> |1f3fe 1f3fe|-| c medium-dark
         🏾   aXa: 2 != 3 -> |61 1f3fe|61|-| c medium-dark
         🏾   aaX: 2 != 3 -> |61|61 1f3fe|-| c medium-dark
         🏿    XX: 1 != 2 -> |1f3ff 1f3ff|-| c dark
         🏿   aXa: 2 != 3 -> |61 1f3ff|61|-| c dark
         🏿   aaX: 2 != 3 -> |61|61 1f3ff|-| c dark

    The codepoints that do accept a skin tone are:
    0x0261D, 0x026F9, 0x0270A, 0x0270B, 0x0270C, 0x0270D, 0x1F385, 0x1F3C2, 0x1F3C3, 0x1F3C4,
    0x1F3C7, 0x1F3CA, 0x1F3CB, 0x1F3CC, 0x1F442, 0x1F443, 0x1F446, 0x1F447, 0x1F448, 0x1F449,
    0x1F44A, 0x1F44B, 0x1F44C, 0x1F44D, 0x1F44E, 0x1F44F, 0x1F450, 0x1F466, 0x1F467, 0x1F468,
    0x1F469, 0x1F46B, 0x1F46C, 0x1F46D, 0x1F46E, 0x1F470, 0x1F471, 0x1F472, 0x1F473, 0x1F474,
    0x1F475, 0x1F476, 0x1F477, 0x1F478, 0x1F47C, 0x1F481, 0x1F482, 0x1F483, 0x1F485, 0x1F486,
    0x1F487, 0x1F4AA, 0x1F574, 0x1F575, 0x1F57A, 0x1F590, 0x1F595, 0x1F596, 0x1F645, 0x1F646,
    0x1F647, 0x1F64B, 0x1F64C, 0x1F64D, 0x1F64E, 0x1F64F, 0x1F6A3, 0x1F6B4, 0x1F6B5, 0x1F6B6,
    0x1F6C0, 0x1F6CC, 0x1F90C, 0x1F90F, 0x1F918, 0x1F919, 0x1F91A, 0x1F91B, 0x1F91C, 0x1F91E,
    0x1F91F, 0x1F926, 0x1F930, 0x1F931, 0x1F932, 0x1F933, 0x1F934, 0x1F935, 0x1F936, 0x1F937,
    0x1F938, 0x1F939, 0x1F93D, 0x1F93E, 0x1F977, 0x1F9B5, 0x1F9B6, 0x1F9B8, 0x1F9B9, 0x1F9BB,
    0x1F9CD, 0x1F9CE, 0x1F9CF, 0x1F9D1, 0x1F9D2, 0x1F9D3, 0x1F9D4, 0x1F9D5, 0x1F9D6, 0x1F9D7,
    0x1F9D8, 0x1F9D9, 0x1F9DA, 0x1F9DB, 0x1F9DC, 0x1F9DD

    """
    it = iter(sorted(data))
    last_item = next(it)
    current_group = [last_item]
    result = [current_group]
    for i in it:
        if i - last_item > max_diff:
            current_group = []
            result.append(current_group)
        current_group.append(i)
        last_item = i
    print('\n'.join(f'{len(g)}:|' + ' '.join(hex(x).replace('0x', '') for x in g) + '|'
                    for g in result))


if __name__ == '__main__':
    parser, run = toolkit('Tests the grapheme break implementation against some unicode version.')
    parser.add_argument('uver', type=float, nargs='?', help='the unicode version to be used')
    parser.add_argument('--all', dest='show_all', action='store_true',
                        help='shows the correct cases, in addition to the wrong ones')
    parser.add_argument('--no-cache', dest='cache', action='store_false',
                        help='ignores the cache and re-downloads the spec')

    run(validate_unicode_breaks)



import argparse

from .. import __version__


def toolkit(descr):
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('-v', '--version', action='version',
                        version=f'alive_progress {__version__}')

    def run(f):
        try:
            f(**parser.parse_args().__dict__)
        except KeyboardInterrupt:
            pass

    return parser, run



"""
Tools intended to be used inside a Python REPL, like ipython or python itself.
"""


def print_chars(line_length=32, max_char=0x20000):
    """Print all chars in the terminal, to help you find that cool one to put in your
    customized spinner or bar. Also useful to determine if your terminal do support them.

    Args:
        line_length (int): the desired characters per line
        max_char (int): the last character in the unicode table to show
            this goes up to 0x10ffff, but after the default value it seems to return
            only question marks, increase it if would like to see more.

    """
    max_char = min(0x10ffff, max(0, max_char))
    for i in range(0x20, max_char + line_length, line_length):
        print(f'0x{i:05x}', end=': ')
        for j in range(line_length):
            if j & 0xf == 0:
                print(' ', end='')
            try:
                print(chr(i + j), end=' ')
            except UnicodeEncodeError:
                print('?', end=' ')
        print()



import os
import sys
from collections import namedtuple
from string import Formatter
from types import FunctionType

from ..utils.terminal import FULL, NON_TTY

ERROR = object()  # represents a config value not accepted.


def _spinner_input_factory(default):
    from ..animations import spinner_compiler
    from ..styles.internal import SPINNERS
    return __style_input_factory(SPINNERS, spinner_compiler,
                                 'spinner_compiler_dispatcher_factory', default)


def _bar_input_factory():
    from ..animations import bars
    from ..styles.internal import BARS
    return __style_input_factory(BARS, bars, 'bar_assembler_factory', None)


def __style_input_factory(name_lookup, module_lookup, inner_name, default):
    def _input(x):
        return name_lookup(x) or func_lookup(x) or default

    name_lookup = __name_lookup_factory(name_lookup)
    func_lookup = __func_lookup_factory(module_lookup, inner_name)
    return _input


def __name_lookup_factory(name_lookup):
    def _input(x):
        if isinstance(x, str):
            return name_lookup.get(x) or ERROR

    return _input


def __func_lookup_factory(module_lookup, inner_name):
    def _input(x):
        if isinstance(x, FunctionType):
            func_file, _ = os.path.splitext(module_lookup.__file__)
            if x.__code__.co_name == inner_name \
                    and os.path.splitext(x.__code__.co_filename)[0] == func_file:
                return x
            return ERROR

    return _input


def _int_input_factory(lower, upper):
    def _input(x):
        if isinstance(x, int) and lower <= x <= upper:
            return int(x)
        return ERROR

    return _input


def _bool_input_factory():
    def _input(x):
        return bool(x)

    return _input


def _force_tty_input_factory():
    def _input(x):
        return table.get(x, ERROR)

    table = {
        None: FULL if sys.stdout.isatty() else NON_TTY,
        False: NON_TTY,
        True: FULL,
    }
    return _input


def _text_input_factory():
    def _input(x):
        return None if x is None else str(x)

    return _input


def _format_input_factory(allowed):
    def _input(x):
        if not isinstance(x, str):
            return bool(x)
        fvars = parser.parse(x)
        if any(f[1] not in allowed for f in fvars):
            # f is a tuple (literal_text, field_name, format_spec, conversion)
            return ERROR
        return x

    # I want to accept only some field names, and pure text.
    allowed = set(allowed.split() + [None])
    parser = Formatter()
    return _input


Config = namedtuple('Config', 'title length spinner bar unknown force_tty disable manual '
                              'enrich_print receipt receipt_text monitor elapsed stats '
                              'title_length spinner_length refresh_secs monitor_end '
                              'elapsed_end stats_end ctrl_c dual_line')


def create_config():
    def reset():
        """Resets global configuration to the default one."""
        set_global(  # this must have all available config vars.
            title=None,
            length=40,
            theme='smooth',  # includes spinner, bar and unknown.
            force_tty=None,
            disable=False,
            manual=False,
            enrich_print=True,
            receipt=True,
            receipt_text=False,
            monitor=True,
            elapsed=True,
            stats=True,
            monitor_end=True,
            elapsed_end=True,
            stats_end=True,
            title_length=0,
            spinner_length=0,
            refresh_secs=0,
            ctrl_c=True,
            dual_line=False,
        )

    def set_global(theme=None, **options):
        """Update the global configuration, to be used in subsequent alive bars.

        See Also:
            alive_progress#alive_bar(**options)

        """
        lazy_init()
        global_config.update(_parse(theme, options))

    def create_context(theme=None, **options):
        """Create an immutable copy of the current configuration, with optional customization."""
        lazy_init()
        local_config = {**global_config, **_parse(theme, options)}
        return Config(**local_config)

    def _parse(theme, options):
        """Validate and convert some configuration options."""

        def validator(key, value):
            try:
                result = validations[key](value)
                if result is ERROR:
                    raise ValueError
                return result
            except KeyError:
                raise ValueError(f'invalid config name: {key}')
            except Exception:
                raise ValueError(f'invalid config value: {key}={value!r}')

        from ..styles.internal import THEMES
        if theme:
            if theme not in THEMES:
                raise ValueError(f'invalid theme name={theme}')
            swap = options
            options = dict(THEMES[theme])
            options.update(swap)
        return {k: validator(k, v) for k, v in options.items()}

    def lazy_init():
        if validations:
            return

        validations.update(  # the ones the user can configure.
            title=_text_input_factory(),
            length=_int_input_factory(3, 300),
            spinner=_spinner_input_factory(None),  # accept empty.
            bar=_bar_input_factory(),
            unknown=_spinner_input_factory(ERROR),  # do not accept empty.
            force_tty=_force_tty_input_factory(),
            disable=_bool_input_factory(),
            manual=_bool_input_factory(),
            enrich_print=_bool_input_factory(),
            receipt=_bool_input_factory(),
            receipt_text=_bool_input_factory(),
            monitor=_format_input_factory('count total percent'),
            monitor_end=_format_input_factory('count total percent'),
            elapsed=_format_input_factory('elapsed'),
            elapsed_end=_format_input_factory('elapsed'),
            stats=_format_input_factory('rate eta'),
            stats_end=_format_input_factory('rate'),
            title_length=_int_input_factory(0, 100),
            spinner_length=_int_input_factory(0, 100),
            refresh_secs=_int_input_factory(0, 60 * 60 * 24),  # maximum 24 hours.
            ctrl_c=_bool_input_factory(),
            dual_line=_bool_input_factory(),
            # title_effect=_enum_input_factory(),  # TODO someday.
        )
        assert all(k in validations for k in Config._fields)  # ensures all fields have validations.

        reset()
        assert all(k in global_config for k in Config._fields)  # ensures all fields have been set.

    global_config, validations = {}, {}
    create_context.set_global, create_context.reset = set_global, reset
    return create_context


config_handler = create_config()



from .core.configuration import config_handler
from .core.progress import alive_bar, alive_it

VERSION = (2, 4, 0)

__author__ = 'Rogério Sampaio de Almeida'
__email__ = 'rsalmei@gmail.com'
__version__ = '.'.join(map(str, VERSION))
__description__ = 'A new kind of Progress Bar, with real-time throughput, ' \
                  'ETA, and very cool animations!'

__all__ = ('alive_bar', 'alive_it', 'config_handler')



"""
Implements support for grapheme clusters and cells (columns on screen).
Graphemes are sequences of codepoints, which are interpreted together based on the Unicode
standard. Grapheme clusters are sequences of graphemes, glued together by Zero Width Joiners.
These graphemes may occupy one or two cells on screen, depending on their glyph size.

Support for these cool chars, like Emojis 😃, was so damn hard to implement because:
1. Python don't know chars that occupy two columns on screen, nor grapheme clusters that are
    rendered as a single char (wide or not), it only understands codepoints;
2. Alive-progress needs to visually align all frames, to keep its progress bars' length from
    popping up and down while running. For this I must somehow know which chars are wide and
    counterbalance that;
3. Alive-progress also has all kinds of animations, which to be generated needs several operations,
    namely len, iterating, indexing, slicing, concatenating and reversing, which now must support
    graphemes and cells! Argh.
4. For that I needed to parse them myself, which I tried but soon realized it was tricky and
    finicky, in addition to changing every year;
5. Then I looked into some lib dependencies, tested several, created the validate tool to help me
    test some Unicode versions, and chose one lib to use;
6. I finally implemented the operations I needed, to the best of my current knowledge, but it
    still wouldn't work. So I tried several spinners to check their alignments, until I finally
    realized what was wrong: I actually needed to align cells, not lengths nor even graphemes!

    For example:
        string \\ length  python  graphemes  cells
             nonono          6        6        6
             🏴󠁧󠁢󠁥󠁮󠁧󠁿👉🏾🏴󠁧󠁢󠁥󠁮󠁧󠁿          16       3        6

7. With that knowledge, I implemented "wide" marks on graphemes, and refactored all operations,
    but it still didn't work. I realized that animations would make these wide chars dynamically
    enter and leave the stage at will, so the frames would end up with different sizes!
    I needed something that could "see" all the frames at once, so I could equalize their sizes...
    So I created the cool spinner compiler, including a light and rocket fast runner;
8. Then I refactored the frame spinner factory, the first and simplest one, and WOW, it worked!
9. To make the others work too, I created the check tool to help me see a spinner's contents,
    directly from the compiled data;
10. The spinner compiler has enabled several improvements in the spinners code, since it ended up
    being a central command center with a lot of functionality, like reshaping and transposing
    the cycle data, or randomizing its playing. The concept of styling parameters and operational
    parameters got stronger with new operational commands, which enabled simpler compound
    animations, without any code duplication. That has enabled me to create the new sequential and
    alongside spinners, way more advanced than before, with intermix and pivot control of cycles;
11. After all of them was working, it was time for the bars, title, exhibit and alive_bar rendering
    itself, which needed to learn how to use the new architecture: change ordinary strings for
    tuples of cells (marked graphemes). All of them needed this same support for the soft wrap to
    keep working;
12. Profit! Only no... But what a ride! 😅

"""

import re
import unicodedata

from .terminal import FULL

PATTERN_SANITIZE = re.compile(r'[\r\n]')
VS_15 = '\ufe0e'


def print_cells(fragments, cols, last_line_len=0, _term=FULL):
    """Print a tuple of fragments of tuples of cells on the terminal, until a given number of
    cols is achieved, slicing over cells when needed.

    Args:
        fragments (Tuple[Union[str, Tuple[str, ...]]): the fragments of message, which are
            joined and gain spaces between them
        cols (int): maximum columns to use
        last_line_len (int): if the size of these fragments are smaller than this, the line is
            cleared before printing anything
        _term: the terminal to be used

    Returns:
        the number of actually used cols.

    """
    available = cols
    _term.write(_term.carriage_return)
    for fragment in filter(None, fragments):
        if fragment == '\n':
            _term.clear_end_line(available)
            available = cols
        elif available == 0:
            continue
        else:
            length = len(fragment)
            if length <= available:
                available -= length
            else:
                available, fragment = 0, fix_cells(fragment[:available])

        _term.write(join_cells(fragment))

    if last_line_len and cols - available < last_line_len:
        _term.clear_end_line(available)

    return cols - available


def join_cells(fragment):
    """Beware, this looses the cell information, converting to a simple string again.
    Don't use unless it is a special case."""
    return ''.join(strip_marks(fragment))


def combine_cells(*fragments):
    """Combine several fragments of cells into one.
    Remember that the fragments get a space between them, so this is mainly to avoid it when
    not desired."""
    return sum(fragments, ())  # this is way faster than tuple(chain.from_iterable()).


def is_wide(g):
    """Try to detect wide chars.

    This is tricky, I've seen several graphemes that have Neutral width (and thus use one
    cell), but actually render as two cells, like shamrock and heart ☘️❤️.
    I've talked to George Nachman, the creator of iTerm2, which has explained to me [1] the fix
    would be to insert a space after these cases, but I can't possibly know if this
    behavior is spread among all terminals, it probably has to do with the Unicode version too,
    so I'm afraid of fixing it.
    Use the `alive_progress.tools.print_chars` tool, and check the section around `0x1f300`
    for more examples.

    [1]: https://gitlab.com/gnachman/iterm2/-/issues/9185

    Args:
        g (str): the grapheme sequence to be tested

    """
    return g[-1] != VS_15 and (len(g) > 1 or unicodedata.east_asian_width(g) in ('W', 'F'))


def fix_cells(chars):
    """Fix truncated cells, removing whole clusters when needed."""
    if not chars:
        return chars

    start = (' ',) if chars[0] is None else ()
    end = (' ',) if chars[-1] is not None and is_wide(chars[-1]) else ()
    return (*start, *chars[bool(start):-1 if end else None], *end)  # noqa


def to_cells(text):
    text = ' '.join(PATTERN_SANITIZE.split(text or ''))
    return mark_graphemes(split_graphemes(text))


def split_graphemes(text):
    from grapheme import graphemes
    return tuple(graphemes(text))


def mark_graphemes(gs):
    return sum(((g, *((None,) if is_wide(g) else ())) for g in gs), ())


def strip_marks(chars):
    return (c for c in chars if c)


def has_wide(text):
    return any(is_wide(x) for x in text)



import sys
from types import SimpleNamespace

from . import tty


def get(original):
    def cols():
        return sys.maxsize  # do not truncate when there's no tty.

    from .void import clear_end_line, clear_end_screen, clear_line  # noqa
    from .void import factory_cursor_up, hide_cursor, show_cursor  # noqa

    flush, write, carriage_return = original.flush, original.write, ''

    return SimpleNamespace(**locals())


BASE = get(tty.BASE)  # support for jupyter notebooks.



import math


def calibrated_fps(calibrate):
    """Calibration of the dynamic frames per second engine.

    I've started with the equation y = log10(x + m) * k + n, where:
      y is the desired fps, m and n are horizontal and vertical translation,
      k is a calibration factor, computed from some user input c (see readme for details).

    Considering minfps and maxfps as given constants, I came to:
      fps = log10(x + 1) * k + minfps, which must be equal to maxfps for x = c,
    so the factor k = (maxfps - minfps) / log10(c + 1), and
      fps = log10(x + 1) * (maxfps - minfps) / log10(c + 1) + minfps

    Neat! ;)

    Args:
        calibrate (float): user provided

    Returns:
        a callable to calculate the fps

    """
    min_fps, max_fps = 2., 60.
    calibrate = max(1e-6, calibrate)
    adjust_log_curve = 100. / min(calibrate, 100.)  # adjust the curve for small numbers
    factor = (max_fps - min_fps) / math.log10((calibrate * adjust_log_curve) + 1.)

    def fps(rate):
        if rate <= 0:
            return 10.  # bootstrap speed
        if rate < calibrate:
            return math.log10((rate * adjust_log_curve) + 1.) * factor + min_fps
        return max_fps

    return fps


def custom_fps(refresh_secs):
    def fps(_rate):
        return refresh_secs

    refresh_secs = 1 / refresh_secs
    return fps



from itertools import chain

import pytest

from alive_progress.animations.utils import extract_fill_graphemes, overlay_sliding_window, \
    round_even, spinner_player, split_options, spread_weighted, static_sliding_window
from alive_progress.utils.cells import join_cells


@pytest.mark.parametrize('gap, contents, length, right, initial, expected_6', [
    (0, ('abc',), 1, False, 0, ('a', 'b', 'c', 'a', 'b', 'c')),
    (0, ('abc',), 1, True, 0, ('a', 'c', 'b', 'a', 'c', 'b')),
    (0, ('abc',), 1, False, -2, ('b', 'c', 'a', 'b', 'c', 'a')),
    (0, ('abc',), 1, True, 1, ('b', 'a', 'c', 'b', 'a', 'c')),
    (0, ('abcdef',), 4, True, 7, ('bcde', 'abcd', 'fabc', 'efab', 'defa', 'cdef')),
    (1, ('abc',), 1, False, 0, ('!', 'a', 'b', 'c', '!', 'a')),
    (1, ('abc',), 1, True, 0, ('!', 'c', 'b', 'a', '!', 'c')),
    (1, ('abc',), 1, False, -2, ('b', 'c', '!', 'a', 'b', 'c')),
    (1, ('abc',), 1, True, 1, ('a', '!', 'c', 'b', 'a', '!')),
    (1, ('abcdef',), 4, True, 7, ('!abc', 'f!ab', 'ef!a', 'def!', 'cdef', 'bcde')),
    (1, ('abc', 'xy'), 1, False, 0, ('!', 'a', 'b', 'c', '!', 'x')),
    (1, ('abc', 'xy'), 1, True, 0, ('!', 'y', 'x', '!', 'c', 'b')),
    (1, ('abc', 'xy'), 1, False, -2, ('x', 'y', '!', 'a', 'b', 'c')),
    (1, ('abc', 'xy'), 1, True, 1, ('a', '!', 'y', 'x', '!', 'c')),
    (1, ('abcdef', 'xy'), 4, True, 7, ('!xy!', 'f!xy', 'ef!x', 'def!', 'cdef', 'bcde')),
])
def test_static_sliding_window(gap, contents, length, right, initial, expected_6):
    ribbon = static_sliding_window('!@#$%', gap, contents, length, right, initial)
    assert tuple(join_cells(next(ribbon)) for _ in range(6)) == expected_6


@pytest.mark.parametrize('gap, contents, length, right, initial, expected_6', [
    (1, ('abcdef',), 4, False, 0, ('!abc', 'abcd', 'bcde', 'cdef', 'def$', 'ef#a')),
    (1, ('abcdef',), 4, True, 0, ('!abc', 'f@ab', 'ef#a', 'def$', 'cdef', 'bcde')),
    (2, ('abc', 'xy'), 2, False, 0, ('!@', '!a', 'ab', 'bc', 'c@', '!@')),
    (3, ('abc', 'xy'), 4, True, 0, ('!@#a', 'y@#$', 'xy#$', '!xy$', '!@xy', '!@#x')),
    (4, ('abc', 'xy'), 6, False, -2, ('xy#$%!', 'y@#$%a', '!@#$ab', '!@#abc', '!@abc!', '!abc%!')),
])
def test_overlay_sliding_window(gap, contents, length, right, initial, expected_6):
    ribbon = overlay_sliding_window('!@#$%', gap, contents, length, right, initial)
    assert tuple(join_cells(next(ribbon)) for _ in range(6)) == expected_6


def test_sliding_window_error():
    with pytest.raises(AssertionError):
        static_sliding_window('back', 10, ('window that slides',), 100, 1, 0)


def spinner_cycle_123():
    # noinspection PyUnusedLocal
    def inner_factory(length=None):
        def inner_spinner():
            yield from '123'

        return inner_spinner

    return inner_factory


def test_spinner_player():
    player = spinner_player(spinner_cycle_123()())
    assert tuple(next(player) for _ in range(4)) == ('1', '2', '3', '1')


@pytest.mark.parametrize('text, default, expected', [
    (None, '<>', '<>'),
    ('', '<>', '<>'),
    ('a', '---', 'a--'),
    ('ab', '!!!', 'ab!'),
    ('ab', '$$', 'ab'),
    ('abc', '##', 'ab'),
])
def test_extract_exactly_n_chars(text, default, expected):
    assert ''.join(chain.from_iterable(extract_fill_graphemes(text, default))) == expected


SAME = object()


@pytest.mark.parametrize('param, expected', [
    (None, (None, None)),
    ('', ('', '')),
    ('ok', ('ok', 'ok')),

    (('a',), SAME),
    (('a', 'b'), SAME),
    (('a', 'b', 'c'), SAME),
])
def test_split_options(param, expected):
    if expected is SAME:
        expected = param
    assert split_options(param) == expected


@pytest.mark.parametrize('actual_length, naturals, expected', [
    (2, (10,), (2,)),
    (8, (2, 2), (4, 4)),
    (14, (2, 4, 1), (4, 8, 2)),
    (12, (10, 24, 8), (3, 7, 2))
])
def test_spread_weighted(actual_length, naturals, expected):
    assert spread_weighted(actual_length, naturals) == expected


@pytest.mark.parametrize('num, expected', [
    (2, 2),
    (3, 4),
    (2.9, 2),
    (3.1, 4),
])
def test_round_even(num, expected):
    assert round_even(num) == expected



import math
from itertools import chain

from .spinner_compiler import spinner_controller
from .utils import combinations, overlay_sliding_window, round_even, spinner_player, \
    split_options, spread_weighted, static_sliding_window
from ..utils.cells import combine_cells, fix_cells, has_wide, mark_graphemes, strip_marks, to_cells


def frame_spinner_factory(*frames):
    """Create a factory of a spinner that delivers frames in sequence, split by cycles.
    Supports unicode grapheme clusters and emoji chars (those that has length one but when on
    screen occupies two cells), as well as all other spinners.

    Args:
        frames (Union[str, Tuple[str, ...]): the frames to be displayed, split by cycles
            if sent only a string, it is interpreted as frames of a single char each.

    Returns:
        a styled spinner factory

    Examples:
        To define one cycle:
        >>> frame_spinner_factory(('cool',))  # only one frame.
        >>> frame_spinner_factory(('ooo', '---'))  # two frames.
        >>> frame_spinner_factory('|/_')  # three frames of one char each, same as below.
        >>> frame_spinner_factory(('|', '/', '_'))

        To define two cycles:
        >>> frame_spinner_factory(('super',), ('cool',))  # one frame each.
        >>> frame_spinner_factory(('ooo', '-'), ('vvv', '^'))  # two frames each.
        >>> frame_spinner_factory('|/_', '▁▄█')  # three frames each, same as below.
        >>> frame_spinner_factory(('|', '/', '_'), ('▁', '▄', '█'))

        Mix and match at will:
        >>> frame_spinner_factory(('oo', '-'), 'cool', ('it', 'is', 'alive!'))

    """
    # shortcut for single char animations.
    frames = (tuple(cycle) if isinstance(cycle, str) else cycle for cycle in frames)

    # support for unicode grapheme clusters and emoji chars.
    frames = tuple(tuple(to_cells(frame) for frame in cycle) for cycle in frames)

    @spinner_controller(natural=max(len(frame) for cycle in frames for frame in cycle))
    def inner_spinner_factory(actual_length=None):
        actual_length = actual_length or inner_spinner_factory.natural
        max_ratio = math.ceil(actual_length / min(len(frame) for cycle in frames
                                                  for frame in cycle))

        def frame_data(cycle):
            for frame in cycle:
                # differently sized frames and repeat support.
                yield (frame * max_ratio)[:actual_length]

        return (frame_data(cycle) for cycle in frames)

    return inner_spinner_factory


def scrolling_spinner_factory(chars, length=None, block=None, background=None, *,
                              right=True, hide=True, wrap=True, overlay=False):
    """Create a factory of a spinner that scrolls characters from one side to
    the other, configurable with various constraints.
    Supports unicode grapheme clusters and emoji chars, those that has length one but when on
    screen occupies two cells.

    Args:
        chars (str): the characters to be scrolled, either together or split in blocks
        length (Optional[int]): the natural length that should be used in the style
        block (Optional[int]): if defined, split chars in blocks with this size
        background (Optional[str]): the pattern to be used besides or underneath the animations
        right (bool): the scroll direction to animate
        hide (bool): controls whether the animation goes through the borders or not
        wrap (bool): makes the animation wrap borders or stop when not hiding.
        overlay (bool): fixes the background in place if overlay, scrolls it otherwise

    Returns:
        a styled spinner factory

    """
    assert not (overlay and not background), 'overlay needs a background'
    assert not (overlay and has_wide(background)), 'unsupported overlay with grapheme background'
    chars, rounder = to_cells(chars), round_even if has_wide(chars) else math.ceil

    @spinner_controller(natural=length or len(chars))
    def inner_spinner_factory(actual_length=None):
        actual_length = actual_length or inner_spinner_factory.natural
        ratio = actual_length / inner_spinner_factory.natural

        initial, block_size = 0, rounder((block or 0) * ratio) or len(chars)
        if hide:
            gap = actual_length
        else:
            gap = max(0, actual_length - block_size)
            if right:
                initial = -block_size if block else abs(actual_length - block_size)

        if block:
            get_block = lambda g: fix_cells((mark_graphemes((g,)) * block_size)[:block_size])
            contents = map(get_block, strip_marks(reversed(chars) if right else chars))
        else:
            contents = (chars,)

        window_impl = overlay_sliding_window if overlay else static_sliding_window
        infinite_ribbon = window_impl(to_cells(background or ' '),
                                      gap, contents, actual_length, right, initial)

        def frame_data():
            for i, fill in zip(range(gap + block_size), infinite_ribbon):
                if i <= size:
                    yield fill

        size = gap + block_size if wrap or hide else abs(actual_length - block_size)
        cycles = len(tuple(strip_marks(chars))) if block else 1
        return (frame_data() for _ in range(cycles))

    return inner_spinner_factory


def bouncing_spinner_factory(chars, length=None, block=None, background=None, *,
                             right=True, hide=True, overlay=False):
    """Create a factory of a spinner that scrolls characters from one side to
    the other and bounce back, configurable with various constraints.
    Supports unicode grapheme clusters and emoji chars, those that has length one but when on
    screen occupies two cells.

    Args:
        chars (Union[str, Tuple[str, str]]): the characters to be scrolled, either
            together or split in blocks. Also accepts a tuple of two strings,
            which are used one in each direction.
        length (Optional[int]): the natural length that should be used in the style
        block (Union[int, Tuple[int, int], None]): if defined, split chars in blocks
        background (Optional[str]): the pattern to be used besides or underneath the animations
        right (bool): the scroll direction to start the animation
        hide (bool): controls whether the animation goes through the borders or not
        overlay (bool): fixes the background in place if overlay, scrolls it otherwise

    Returns:
        a styled spinner factory

    """
    chars_1, chars_2 = split_options(chars)
    block_1, block_2 = split_options(block)
    scroll_1 = scrolling_spinner_factory(chars_1, length, block_1, background, right=right,
                                         hide=hide, wrap=False, overlay=overlay)
    scroll_2 = scrolling_spinner_factory(chars_2, length, block_2, background, right=not right,
                                         hide=hide, wrap=False, overlay=overlay)
    return sequential_spinner_factory(scroll_1, scroll_2)


def sequential_spinner_factory(*spinner_factories, intermix=True):
    """Create a factory of a spinner that combines other spinners together, playing them
    one at a time sequentially, either intermixing their cycles or until depletion.

    Args:
        spinner_factories (spinner): the spinners to be combined
        intermix (bool): intermixes the cycles if True, generating all possible combinations;
            runs each one until depletion otherwise.

    Returns:
        a styled spinner factory

    """

    @spinner_controller(natural=max(factory.natural for factory in spinner_factories))
    def inner_spinner_factory(actual_length=None):
        actual_length = actual_length or inner_spinner_factory.natural
        spinners = [factory(actual_length) for factory in spinner_factories]

        def frame_data(spinner):
            yield from spinner()

        if intermix:
            cycles = combinations(spinner.cycles for spinner in spinners)
            gen = ((frame_data(spinner) for spinner in spinners)
                   for _ in range(cycles))
        else:
            gen = ((frame_data(spinner) for _ in range(spinner.cycles))
                   for spinner in spinners)

        return (c for c in chain.from_iterable(gen))  # transforms the chain to a gen exp.

    return inner_spinner_factory


def alongside_spinner_factory(*spinner_factories, pivot=None):
    """Create a factory of a spinner that combines other spinners together, playing them
    alongside simultaneously. Each one uses its own natural length, which is spread weighted
    to the available space.

    Args:
        spinner_factories (spinner): the spinners to be combined
        pivot (Optional[int]): the index of the spinner to dictate the animation cycles
            if None, the whole animation will be compiled into a unique cycle.

    Returns:
        a styled spinner factory

    """

    @spinner_controller(natural=sum(factory.natural for factory in spinner_factories))
    def inner_spinner_factory(actual_length=None, offset=0):
        if actual_length:
            lengths = spread_weighted(actual_length, [f.natural for f in spinner_factories])
            actual_pivot = None if pivot is None or not lengths[pivot] \
                else spinner_factories[pivot](lengths[pivot])
            spinners = [factory(length) for factory, length in
                        zip(spinner_factories, lengths) if length]
        else:
            actual_pivot = None if pivot is None else spinner_factories[pivot]()
            spinners = [factory() for factory in spinner_factories]

        def frame_data(cycle_gen):
            yield from (combine_cells(*fragments) for _, *fragments in cycle_gen)

        frames = combinations(spinner.total_frames for spinner in spinners)
        spinners = [spinner_player(spinner) for spinner in spinners]
        [[next(player) for _ in range(i * offset)] for i, player in enumerate(spinners)]

        if actual_pivot is None:
            breaker, cycles = lambda: range(frames), 1
        else:
            breaker, cycles = lambda: actual_pivot(), \
                              frames // actual_pivot.total_frames * actual_pivot.cycles
        return (frame_data(zip(breaker(), *spinners)) for _ in range(cycles))

    return inner_spinner_factory


def delayed_spinner_factory(spinner_factory, copies, offset=1, *, dynamic=True):
    """Create a factory of a spinner that combines itself several times alongside,
    with an increasing iteration offset on each one.

    Args:
        spinner_factory (spinner): the source spinner
        copies (int): the number of copies
        offset (int): the offset to be applied incrementally to each copy
        dynamic (bool): dynamically changes the number of copies based on available space

    Returns:
        a styled spinner factory

    """

    if not dynamic:
        factories = (spinner_factory,) * copies
        return alongside_spinner_factory(*factories, pivot=0).op(offset=offset)

    @spinner_controller(natural=spinner_factory.natural * copies, skip_compiler=True)
    def inner_spinner_factory(actual_length=None):
        n = math.ceil(actual_length / spinner_factory.natural) if actual_length else copies
        return delayed_spinner_factory(spinner_factory, n, offset, dynamic=False)(actual_length)

    return inner_spinner_factory



import pytest

from alive_progress.animations.spinners import alongside_spinner_factory, \
    bouncing_spinner_factory, delayed_spinner_factory, frame_spinner_factory, \
    scrolling_spinner_factory, sequential_spinner_factory
from alive_progress.utils.cells import join_cells


@pytest.mark.parametrize('frames, expected', [
    ('a\nb', (('a', ' ', 'b'),)),
    ('abc', (('a', 'b', 'c'),)),
    (('a\nb', '\nc '), (('a b', ' c '),)),
    (('a  ', ' b ', '  c'), (('a  ', ' b ', '  c'),)),
    (('a', '(a)', ' (*) '), (('aaaaa', '(a)(a', ' (*) '),)),
    (('ok', '😺😺'), (('okok', '😺😺'),)),
])
def test_frame_spinner(frames, expected):
    spinner_factory = frame_spinner_factory(frames)
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('length, block, background, right, hide, expected', [
    (None, None, ' ', True, True, (('   ', 'c  ', 'bc ', 'abc', ' ab', '  a'),)),
    (None, None, ' ', False, True, (('   ', '  a', ' ab', 'abc', 'bc ', 'c  '),)),
    (None, None, ' ', True, False, (('abc', 'cab', 'bca'),)),
    (None, None, ' ', False, False, (('abc', 'bca', 'cab'),)),
    (2, None, '~', True, True, (('~~', 'c~', 'bc', 'ab', '~a'),)),
    (2, None, '~', True, False, (('bc', 'ab', 'ca'),)),
    (2, None, '~', False, True, (('~~', '~a', 'ab', 'bc', 'c~'),)),
    (2, None, '~', False, False, (('ab', 'bc', 'ca'),)),
    (3, None, '~', True, True, (('~~~', 'c~~', 'bc~', 'abc', '~ab', '~~a'),)),
    (3, None, '~', True, False, (('abc', 'cab', 'bca'),)),
    (3, None, '~', False, True, (('~~~', '~~a', '~ab', 'abc', 'bc~', 'c~~'),)),
    (3, None, '~', False, False, (('abc', 'bca', 'cab'),)),
    (4, None, ' ', True, True, (('    ', 'c   ', 'bc  ', 'abc ', ' abc', '  ab', '   a'),)),
    (4, None, ' ', True, False, (('abc ', ' abc', 'c ab', 'bc a'),)),
    (4, None, ' ', False, True, (('    ', '   a', '  ab', ' abc', 'abc ', 'bc  ', 'c   '),)),
    (4, None, ' ', False, False, ((' abc', 'abc ', 'bc a', 'c ab'),)),
    (4, 1, '_', True, True, (('____', 'a___', '_a__', '__a_', '___a'),
                             ('____', 'b___', '_b__', '__b_', '___b'),
                             ('____', 'c___', '_c__', '__c_', '___c'))),
    (4, 2, '_', True, False, (('aa__', '_aa_', '__aa', 'b__a'),
                              ('bb__', '_bb_', '__bb', 'c__b'),
                              ('cc__', '_cc_', '__cc', 'a__c'))),
])
def test_scrolling_spinner(length, block, background, right, hide, expected):
    spinner_factory = scrolling_spinner_factory('abc', length, block, background,
                                                right=right, hide=hide)
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('length, block, background, hide, expected', [
    (None, None, None, True, (('   ', 'c  ', 'bc ', 'abc', ' ab', '  a'),
                              ('   ', '  d', ' de', 'def', 'ef ', 'f  '),)),
    (None, None, None, False, (('abc',), ('def',),)),
    (2, None, '~', True, (('~~', 'c~', 'bc', 'ab', '~a'), ('~~', '~d', 'de', 'ef', 'f~'),)),
    (2, None, '~', False, (('bc', 'ab'), ('de', 'ef'),)),
    (3, None, '+', True, (('+++', 'c++', 'bc+', 'abc', '+ab', '++a'),
                          ('+++', '++d', '+de', 'def', 'ef+', 'f++'),)),
    (3, None, '+', False, (('abc',), ('def',),)),
    (4, None, ' ', True, (('    ', 'c   ', 'bc  ', 'abc ', ' abc', '  ab', '   a'),
                          ('    ', '   d', '  de', ' def', 'def ', 'ef  ', 'f   '),)),
    (4, None, ' ', False, (('abc ', ' abc'), (' def', 'def '),)),
    (3, 1, '_', True, (('___', 'a__', '_a_', '__a'),
                       ('___', '__d', '_d_', 'd__'),
                       ('___', 'b__', '_b_', '__b'),
                       ('___', '__e', '_e_', 'e__'),
                       ('___', 'c__', '_c_', '__c'),
                       ('___', '__f', '_f_', 'f__'))),
    (5, 2, '_', False, (('aa___', '_aa__', '__aa_', '___aa'),
                        ('___dd', '__dd_', '_dd__', 'dd___'),
                        ('bb___', '_bb__', '__bb_', '___bb'),
                        ('___ee', '__ee_', '_ee__', 'ee___'),
                        ('cc___', '_cc__', '__cc_', '___cc'),
                        ('___ff', '__ff_', '_ff__', 'ff___'))),
])
def test_bouncing_spinner(length, block, background, hide, expected):
    spinner_factory = bouncing_spinner_factory(('abc', 'def'), length, block, background,
                                               right=True, hide=hide)
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('inputs, expected', [
    (('123', 'abc'), (('1a', '2b', '3c'),)),
    (('12', 'abc'), (('1a', '2b', '1c', '2a', '1b', '2c'),)),
    ((('12', '34', '56'), 'ab'), (('12a', '34b', '56a', '12b', '34a', '56b'),)),
])
def test_alongside_spinner(inputs, expected, spinner_test):
    spinner_factory = alongside_spinner_factory(*(spinner_test(x) for x in inputs))
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('inputs, expected', [
    (('123', 'abc'), (('1a', '2b', '3c'),)),
    (('12', 'abc'), (('1a', '2b'), ('1c', '2a'), ('1b', '2c'))),
    ((('12', '34', '56'), 'ab'), (('12a', '34b', '56a'), ('12b', '34a', '56b'))),
])
def test_alongside_spinner_with_pivot(inputs, expected, spinner_test):
    spinner_factory = alongside_spinner_factory(*(spinner_test(x) for x in inputs), pivot=0)
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('inputs, expected', [
    (('123', 'abc'), (('11a', '22b', '33c'),)),
    (('12', 'abc'), (('11a', '22b', '11c', '22a', '11b', '22c'),)),
    ((('12', '34', '56'), 'ab'), (('12a', '34b', '56a', '12b', '34a', '56b'),)),
])
def test_alongside_spinner_custom(inputs, expected, spinner_test):
    spinner_factory = alongside_spinner_factory(*(spinner_test(x) for x in inputs))
    spinner = spinner_factory(3)  # custom spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('inputs, expected', [
    (('123', 'abc'), (('1',), ('a',), ('2',), ('b',), ('3',), ('c',))),
    (('12', 'abc'), (('1',), ('a',), ('2',), ('b',), ('1',), ('c',),
                     ('2',), ('a',), ('1',), ('b',), ('2',), ('c',))),
    ((('12', '34', '56'), 'ab'), (('1', '2'), ('a',), ('3', '4'), ('b',), ('5', '6'), ('a',),
                                  ('1', '2'), ('b',), ('3', '4'), ('a',), ('5', '6'), ('b',))),
])
def test_sequential_spinner(inputs, expected, spinner_test):
    spinner_factory = sequential_spinner_factory(*(spinner_test(*x) for x in inputs))
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('inputs, expected', [
    (('123', 'abc'), (('1',), ('2',), ('3',), ('a',), ('b',), ('c',))),
    (('12', 'abc'), (('1',), ('2',), ('a',), ('b',), ('c',))),
    ((('12', '34', '56'), 'ab'), (('1', '2'), ('3', '4'), ('5', '6'), ('a',), ('b',))),
])
def test_sequential_spinner_no_intermix(inputs, expected, spinner_test):
    spinner_factory = sequential_spinner_factory(*(spinner_test(*x) for x in inputs),
                                                 intermix=False)
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected


@pytest.mark.parametrize('copies, offset, expected', [
    (3, 1, (('123', '234', '345', '451', '512'),)),
    (4, 2, (('1352', '2413', '3524', '4135', '5241'),)),
])
def test_delayed_spinner(copies, offset, expected, spinner_test):
    spinner_factory = delayed_spinner_factory(spinner_test('12345'), copies, offset)
    spinner = spinner_factory()  # natural spinner size.
    assert tuple(tuple(join_cells(f) for f in spinner()) for _ in expected) == expected



import sys
from types import SimpleNamespace

from . import jupyter, non_tty, tty, void

# work around a bug on Windows' command prompt, where ANSI escape codes are disabled by default.
if sys.platform == 'win32':
    import os

    os.system('')


def _create(mod, interactive):
    terminal = SimpleNamespace(
        interactive=interactive,
        cursor_up_1=mod.factory_cursor_up(1),

        # directly from terminal impl.
        write=mod.write,
        flush=mod.flush,
        cols=mod.cols,
        carriage_return=mod.carriage_return,
        clear_line=mod.clear_line,
        clear_end_line=mod.clear_end_line,
        clear_end_screen=mod.clear_end_screen,
        hide_cursor=mod.hide_cursor,
        show_cursor=mod.show_cursor,
        factory_cursor_up=mod.factory_cursor_up,
    )
    return terminal


def _is_notebook():
    """This detection is tricky, because by design there's no way to tell which kind
    of frontend is connected, there may even be more than one with different types!
    Also, there may be other types I'm not aware of...
    So, I've chosen what I thought it was the safest method, with a negative logic:
    if it _isn't_ None or TerminalInteractiveShell, it should be the "jupyter" type.
    The jupyter type does not emit any ANSI Escape Codes.
    """
    if 'IPython' not in sys.modules:
        # if IPython hasn't been imported, there's nothing to check.
        return False

    # noinspection PyPackageRequirements
    from IPython import get_ipython
    class_ = get_ipython().__class__.__name__
    return class_ != 'TerminalInteractiveShell'


FULL = _create(jupyter.BASE if _is_notebook() else tty.BASE, True)
NON_TTY = _create(non_tty.BASE, False)
VOID = _create(void, False)



from distutils.core import setup

from setuptools import find_packages

import alive_progress


def get_readme():
    with open('README.md', encoding='utf-8') as readme_file:
        return readme_file.read()


setup(
    name='alive-progress',
    version=alive_progress.__version__,
    description=alive_progress.__description__,
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/rsalmei/alive-progress',
    author=alive_progress.__author__,
    author_email=alive_progress.__email__,
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Natural Language :: English',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.2',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        # 'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    keywords='progress bar progress-bar progressbar spinner eta monitoring python terminal '
             'multi-threaded REPL alive animated visual feedback simple live efficient'.split(),
    packages=find_packages(),
    data_files=[('', ['LICENSE'])],
    python_requires='>=3.7, <4',
    install_requires=['about_time==3.1.1', 'grapheme==0.6.0'],
)



import timeit

from about_time import duration_human

from .utils import toolkit
from ..core.configuration import config_handler
from ..core.progress import __alive_bar


def overhead(total=None, *, calibrate=None, **options):
    number = 400  # timeit number of runs inside each repetition.
    repeat = 300  # timeit how many times to repeat the whole test.

    config = config_handler(disable=True, **options)
    with __alive_bar(config, total, calibrate=calibrate, _cond=__lock, _sampling=True) as loc:
        # the timing of the print_cells function increases proportionately with the
        # number of columns in the terminal, so I want a baseline here `VOID.cols == 0`.
        res = timeit.repeat('alive_repr()', repeat=repeat, number=number, globals=loc)

    return duration_human(min(res) / number).replace('us', 'µs')


OVERHEAD_SAMPLING_GROUP = [
    ('definite', dict(total=1)),
    ('manual(b)', dict(total=1, manual=True)),
    ('manual(u)', dict(manual=True)),
    ('unknown', dict()),
]
OVERHEAD_SAMPLING = [
    ('default', dict()),
    ('receipt', dict(receipt_text=True)),
    ('no spinner', dict(spinner=None)),
    ('no elapsed', dict(elapsed=False)),
    ('no monitor', dict(monitor=False)),
    ('no stats', dict(stats=False)),
    ('no bar', dict(bar=None)),
    ('only spinner', dict(bar=None, monitor=False, elapsed=False, stats=False)),
    ('only elapsed', dict(bar=None, spinner=None, monitor=False, stats=False)),
    ('only monitor', dict(bar=None, spinner=None, elapsed=False, stats=False)),
    ('only stats', dict(bar=None, spinner=None, monitor=False, elapsed=False)),
    ('only bar', dict(spinner=None, monitor=False, elapsed=False, stats=False)),
    ('none', dict(bar=None, spinner=None, monitor=False, elapsed=False, stats=False)),
]


def overhead_sampling():
    print('warmup', end='', flush=True)
    for _ in range(5):
        print('.', end='', flush=True)
        overhead()
    print('\r', end='', flush=True)

    max_name = max(len(x) for x, _ in OVERHEAD_SAMPLING)
    print(f'{"":>{max_name}} | {" | ".join(g for g, _ in OVERHEAD_SAMPLING_GROUP)} |')
    for name, config in OVERHEAD_SAMPLING:
        print(f'{name:>{max_name}} ', end='', flush=True)
        for group, data in OVERHEAD_SAMPLING_GROUP:
            print(f'| {overhead(**data, **config):^{len(group)}} ', end='', flush=True)
        print('|')


def __noop_p(_ignore):
    return 0


class __lock:
    def __enter__(self):
        pass

    def __exit__(self, _type, value, traceback):
        pass


if __name__ == '__main__':
    parser, run = toolkit('Estimates the alive_progress overhead per cycle on your system.')

    run(overhead_sampling)



"""
Very basic color implementation, just to print fixed messages on screen.

It's very hard to support colors inside any moving parts of alive-progress, as I would need
to implement operations like len, indexing, slicing, concatenation and revert, while maintaining
the color information correct! There's an impedance mismatch between what we see and how we
represent it (yeah, similar to unicode grapheme clusters 😅).

I'd say it is possible now with the new cell architecture, I'd just need some new rules.
I would need to split color information just like grapheme clusters and mark them like cells,
but with the opposite effect. In cell the marks increase the len for the wide chars, but here
they must not change it somehow.
Also, the ansi escape codes the terminal receives changes its state, so I couldn't simply truncate
the line anymore, it'd need another escape code to finalize that one, and return it to its previous
state!
So, these special cells would require yet more work.

"""


def color_factory(color_code):
    def apply(text: str, format_spec=''):
        return f'{color_code}{text:{format_spec}}\x1b[0m'

    def mix(*colors):
        return [color_factory(c.color_code + color_code) for c in colors]

    apply.mix, apply.color_code = mix, color_code
    return apply


# colors
BLUE = color_factory('\x1b[94m')
GREEN = color_factory('\x1b[92m')
YELLOW = color_factory('\x1b[93m')
RED = color_factory('\x1b[91m')
MAGENTA = color_factory('\x1b[95m')
CYAN = color_factory('\x1b[96m')
ORANGE = color_factory('\x1b[38;5;208m')

# modifiers
BOLD = color_factory('\x1b[1m')
DIM = color_factory('\x1b[2m')
ITALIC = color_factory('\x1b[3m')
UNDERLINE = color_factory('\x1b[4m')

BLUE_BOLD, BLUE_DIM, BLUE_IT, BLUE_UNDER = BLUE.mix(BOLD, DIM, ITALIC, UNDERLINE)
GREEN_BOLD, GREEN_DIM, GREEN_IT, GREEN_UNDER = GREEN.mix(BOLD, DIM, ITALIC, UNDERLINE)
YELLOW_BOLD, YELLOW_DIM, YELLOW_IT, YELLOW_UNDER = YELLOW.mix(BOLD, DIM, ITALIC, UNDERLINE)
RED_BOLD, RED_DIM, RED_IT, RED_UNDER = RED.mix(BOLD, DIM, ITALIC, UNDERLINE)
MAGENTA_BOLD, MAGENTA_DIM, MAGENTA_IT, MAGENTA_UNDER = MAGENTA.mix(BOLD, DIM, ITALIC, UNDERLINE)
CYAN_BOLD, CYAN_DIM, CYAN_IT, CYAN_UNDER = CYAN.mix(BOLD, DIM, ITALIC, UNDERLINE)
ORANGE_BOLD, ORANGE_DIM, ORANGE_IT, ORANGE_UNDER = ORANGE.mix(BOLD, DIM, ITALIC, UNDERLINE)



from ..animations.bars import bar_factory
from ..animations.spinners import alongside_spinner_factory, bouncing_spinner_factory, \
    delayed_spinner_factory, frame_spinner_factory, scrolling_spinner_factory, \
    sequential_spinner_factory


def _filter(context):
    return {k: v for k, v in context.items() if not k.startswith('_')}


def __create_spinners():
    classic = frame_spinner_factory(r'-\|/')
    stars = scrolling_spinner_factory('*', 4, 1, hide=False)
    twirl = frame_spinner_factory('←↖↑↗→↘↓↙')
    twirls = delayed_spinner_factory(twirl, 3)
    horizontal = frame_spinner_factory('▏▎▍▌▋▊▉█').reshape(1).bounce().reshape(7)
    vertical = frame_spinner_factory('▁▂▃▄▅▆▇█').reshape(1).bounce().reshape(7)
    waves = delayed_spinner_factory(vertical, 3, 2)
    waves2 = delayed_spinner_factory(vertical, 3, 5)
    waves3 = delayed_spinner_factory(vertical, 3, 7)
    dots = frame_spinner_factory('⠁⠈⠐⠠⢀⡀⠄⠂')
    dots_waves = delayed_spinner_factory(dots, 5)
    dots_waves2 = delayed_spinner_factory(dots, 5, 2)

    _balloon = bouncing_spinner_factory('🎈', 12, background='⠁⠈⠐⠠⢀⡀⠄⠂', overlay=True)
    it = sequential_spinner_factory(  # do not use block mode, so that they doesn't grow.
        _balloon,
        _balloon,  # makes the balloon twice as common.
        bouncing_spinner_factory('🤡', background='⠁⠈⠐⠠⢀⡀⠄⠂', overlay=False),
        intermix=False
    ).randomize()

    ball_belt = bouncing_spinner_factory('●', 8, 0, '< >', hide=False)
    balls_belt = bouncing_spinner_factory('●', 8, 1, r'/~\_', hide=False)
    triangles = bouncing_spinner_factory(('▶', '◀'), 6, 2, hide=False)
    brackets = bouncing_spinner_factory(('>', '<'), 8, 3, hide=False)
    bubbles = bouncing_spinner_factory(('∙●⦿', '○'), 10, 5, hide=False)
    circles = bouncing_spinner_factory('●', 8, background='○', hide=False)
    squares = bouncing_spinner_factory('■', 8, background='□', hide=False)
    flowers = bouncing_spinner_factory('💐🌷🌸🌹🌺🌻🌼', 12, (2, 4)).pause(center=6).randomize()
    elements = bouncing_spinner_factory(('🔥💨', '🌊⚡️'), 6, 2)
    loving = bouncing_spinner_factory(('😍🥰', '⭐️🤩'), 8, (2, 3), '. ', hide=False, overlay=True)

    notes = bouncing_spinner_factory(('♩♪', '♫♬'), 8, 2, hide=False).pause(other=2)
    notes2 = delayed_spinner_factory(scrolling_spinner_factory('♩♪♫♬'), 3)

    arrow = scrolling_spinner_factory('>>----->', 15)
    arrows = bouncing_spinner_factory(('→', '←'), 6, 3)
    arrows2 = scrolling_spinner_factory('→➜➞➣➤➩➪➮', 5, 2, hide=False)
    _arrows_left = scrolling_spinner_factory('.˱·˂°❮', 6, 3, right=False)
    _arrows_right = scrolling_spinner_factory('.˲·˃°❯', 6, 3, right=True)
    arrows_in = alongside_spinner_factory(_arrows_right, _arrows_left)
    arrows_out = alongside_spinner_factory(_arrows_left, _arrows_right)

    _core = frame_spinner_factory('∙○⦿●')
    radioactive = alongside_spinner_factory(_arrows_left, _core, _arrows_right)

    boat = bouncing_spinner_factory((r'*|___/', r'\___|*'), 12, background='_.--.',
                                    hide=False, overlay=True)
    fish = scrolling_spinner_factory("><((('>", 15, hide=False)
    fish2 = bouncing_spinner_factory(("><('>", "<')><"), 12, hide=False)
    _fish_trail = scrolling_spinner_factory('¸.·´¯`·.·´¯`·.¸¸.·´¯`·.><(((º>', 15)
    _small_fishes = bouncing_spinner_factory(('><>     ><>', '<><  <><    <><'), 15)
    fishes = sequential_spinner_factory(_small_fishes, _fish_trail)
    crab = bouncing_spinner_factory((r'Y (••) Y', r'Y (  ) Y'), 15, background='.,.,,..,.,',
                                    hide=False, overlay=True)  # hey it's Ferris #rustacean!

    _look = bouncing_spinner_factory(('Look!', "It's moving!"))
    _alive = bouncing_spinner_factory(("It's alive!", "IT'S ALIVE!!"))
    frank = sequential_spinner_factory(_look, _alive, intermix=False)

    wait = scrolling_spinner_factory('please wait...', right=False)
    wait2 = bouncing_spinner_factory(('please', 'wait'), 15, hide=False).pause()
    wait3 = bouncing_spinner_factory(('processing',
                                      'well, this is taking longer than anticipated, hold on'), 15)
    pulse = frame_spinner_factory((
        r'•––––––––––––', r'•––––––––––––', r'•––––––––––––', r'•––––––––-–––',
        r'–•–––––––––––', r'–•–––––––––––', r'–•–––––––––––', r'–•–––––––––––',
        r'––•––––––––––', r'––√––––––––––', r'––•––––––––––', r'––•––––––––––',
        r'–––•–––––––––', r'––√•–––––––––', r'–––•–––––––––', r'–––•–––––––––',
        r'––––•––––––––', r'––√-•––––––––', r'––––√––––––––', r'––––•––––––––',
        r'–––––•–––––––', r'––√--•–––––––', r'––––√\–––––––', r'–––––•–––––––',
        r'––––––•––––––', r'––√--–•––––––', r'––––√\/––––––', r'––––––•––––––',
        r'–––––––•–––––', r'–––--––•–––––', r'––––√\/•–––––', r'–––––––√–––––',
        r'––––––––•––––', r'––––-–––•––––', r'––––√\/–•––––', r'–––––––√\––––',
        r'–––––––––•–––', r'–––––––––•–––', r'–––––\/––•–––', r'–––––––√\•–––',
        r'––––––––––•––', r'––––––––––•––', r'––––––/–––•––', r'–––––––√\-•––',
        r'–––––––––––•–', r'–––––––––––•–', r'–––––––––––•–', r'–––––––√\-–•–',
        r'––––––––––––•', r'––––––––––––•', r'––––––––––––•', r'––––––––\-––•',
    )).reshape(4).transpose().randomize()

    return _filter(locals())


def __create_bars():
    smooth = bar_factory('▏▎▍▌▋▊▉█')
    classic = bar_factory('=', tip='>', borders='[]', errors='!x')
    classic2 = bar_factory('#', background='.', borders='[]', errors='!x')
    brackets = bar_factory('>')
    blocks = bar_factory('▏▎▍▌▋▊▉')
    bubbles = bar_factory('∙○⦿●', borders='<>')
    solid = bar_factory('∙□☐■', borders='<>')
    checks = bar_factory('✓')
    circles = bar_factory('●', background='○', borders='<>')
    squares = bar_factory('■', background='□', borders='<>')
    halloween = bar_factory('🎃', background='   👻   💀', errors=('😱', '🗡🗡🗡🗡'))
    filling = bar_factory('▁▂▃▄▅▆▇█')
    notes = bar_factory('♩♪♫♬', errors='♭♯')
    ruler = bar_factory(tip='┃', background='∙∙∙∙.')
    ruler2 = bar_factory(tip='┃', background='∙∙∙∙+')
    fish = bar_factory(tip="><('>", background='¸.·´¯`·.·´¯`·.¸¸.·´¯`·.')
    scuba = bar_factory(tip='>=≗)o', background='_)_)._∙__⠈__)○____∙○___)__⠈(_(__')

    return _filter(locals())


def __create_themes():
    smooth = dict(bar='smooth', spinner='waves', unknown='triangles')
    classic = dict(bar='classic', spinner='classic', unknown='brackets')
    scuba = dict(bar='scuba', spinner='fish2', unknown='fishes')  # I love scuba-diving.
    musical = dict(bar='notes', spinner='notes', unknown='notes2')

    return _filter(locals())


SPINNERS = __create_spinners()
BARS = __create_bars()
THEMES = __create_themes()



from unittest import mock

import pytest

from alive_progress.core.progress import _GatedAssignProperty, _GatedProperty, _render_title
from alive_progress.utils.cells import join_cells


@pytest.mark.parametrize('length, text, expected', [
    (0, None, ''),
    (0, '', ''),
    (0, 'c', 'c'),
    (0, 'cool bar title', 'cool bar title'),
    (1, None, ' '),
    (1, '', ' '),
    (1, 'c', 'c'),
    (1, 'cool bar title', '…'),
    (1, '😺', '…'),
    (2, '😺', '😺'),
    (5, 'cool bar title', 'cool…'),
    (14, 'cool bar title', 'cool bar title'),
    (20, 'cool bar title', 'cool bar title      '),
    (15, 'cool bar title😺', 'cool bar title…'),
    (16, 'cool bar title😺', 'cool bar title😺'),
    (16, 'cool bar title😺a', 'cool bar title …'),
    (16, 'cool bar title😺😺', 'cool bar title …'),
])
def test_render_title(length, text, expected):
    local_config = mock.Mock(title=text, title_length=length)
    assert join_cells(_render_title(local_config)) == expected


def test_gated_properties():
    class AClass:
        readonly = _GatedProperty()
        assignable = _GatedAssignProperty()

    instance, m = AClass(), mock.Mock()
    instance._handle = True
    instance._readonly = lambda: 1
    instance._assignable = m

    assert instance.readonly() == 1

    instance.assignable()
    m.assert_called_once_with()

    m.reset_mock()
    instance.assignable(2)
    m.assert_called_once_with(2)

    m.reset_mock()
    instance.assignable = 3
    m.assert_called_once_with(3)



import logging
import sys
from collections import defaultdict
from functools import partial
from itertools import chain, islice, repeat
from logging import StreamHandler
from types import SimpleNamespace

# support for click.echo, which calls `write` with bytes instead of str.
ENCODING = sys.getdefaultencoding()


def buffered_hook_manager(header_template, get_pos, cond_refresh, term):
    """Create and maintain a buffered hook manager, used for instrumenting print
    statements and logging.

    Args:
        header_template (): the template for enriching output
        get_pos (Callable[..., Any]): the container to retrieve the current position
        cond_refresh: Condition object to force a refresh when printing
        term: the current terminal

    Returns:
        a closure with several functions

    """

    def flush_buffers():
        for stream, buffer in buffers.items():
            flush(stream)

    def flush(stream):
        if buffers[stream]:
            write(stream, '\n')
            stream.flush()

    def write(stream, part):
        if isinstance(part, bytes):
            part = part.decode(ENCODING)

        buffer = buffers[stream]
        if part != '\n':
            # this will generate a sequence of lines interspersed with None, which will later
            # be rendered as the indent filler to align additional lines under the same header.
            gen = chain.from_iterable(zip(repeat(None), part.splitlines(True)))
            buffer.extend(islice(gen, 1, None))
        else:
            header = get_header()
            with cond_refresh:
                spacer = ' ' * len(header)
                nested = ''.join(line or spacer for line in buffer)
                text = f'{header}{nested.rstrip()}\n'
                if stream in base:  # pragma: no cover
                    # use the current terminal abstraction for preparing the screen.
                    term.clear_line()
                # handle all streams, both screen and logging.
                stream.write(text)
                stream.flush()
                cond_refresh.notify()
                buffer[:] = []

    def get_hook_for(handler):
        if handler.stream:  # supports FileHandlers with delay=true.
            handler.stream.flush()
        return SimpleNamespace(write=partial(write, handler.stream),
                               flush=partial(flush, handler.stream),
                               isatty=sys.stdout.isatty)

    def install():
        def get_all_loggers():
            yield logging.root
            yield from (logging.getLogger(name) for name in logging.root.manager.loggerDict)

        # modify all stream handlers, including their subclasses.
        before_handlers.update({h: _set_stream(h, get_hook_for(h))  # noqa
                                for logger in get_all_loggers()
                                for h in logger.handlers if isinstance(h, StreamHandler)})
        sys.stdout, sys.stderr = (get_hook_for(SimpleNamespace(stream=x)) for x in base)

    def uninstall():
        flush_buffers()
        buffers.clear()
        sys.stdout, sys.stderr = base

        [_set_stream(handler, original_stream)
         for handler, original_stream in before_handlers.items()]
        before_handlers.clear()

        # does the number of logging handlers changed??
        # if yes, it probably means logging was initialized within alive_bar context,
        # and thus there can be an instrumented stdout or stderr within handlers,
        # which causes a TypeError: unhashable type: 'types.SimpleNamespace'...

    # internal data.
    buffers = defaultdict(list)
    get_header = gen_header(header_template, get_pos) if header_template else null_header
    base = sys.stdout, sys.stderr  # needed for tests.
    before_handlers = {}

    # external interface.
    hook_manager = SimpleNamespace(
        flush_buffers=flush_buffers,
        install=install,
        uninstall=uninstall,
    )

    return hook_manager


def passthrough_hook_manager():  # pragma: no cover
    passthrough_hook_manager.flush_buffers = __noop
    passthrough_hook_manager.install = __noop
    passthrough_hook_manager.uninstall = __noop
    return passthrough_hook_manager


def __noop():  # pragma: no cover
    pass


def gen_header(header_template, get_pos):  # pragma: no cover
    def inner():
        return header_template.format(get_pos())

    return inner


def null_header():  # pragma: no cover
    return ''


if sys.version_info >= (3, 7):  # pragma: no cover
    def _set_stream(handler, stream):
        return handler.setStream(stream)
else:  # pragma: no cover
    def _set_stream(handler, stream):
        # from python 3.7 implementation.
        result = handler.stream
        handler.acquire()
        try:
            handler.flush()
            handler.stream = stream
        finally:
            handler.release()
        return result
